"""Database store implementations
"""
import json

# import psycopg2 to handel database transactions
import psycopg2
import psycopg2.extras
from smif.data_layer.abstract_config_store import ConfigStore
from smif.data_layer.abstract_data_store import DataStore
from smif.data_layer.abstract_metadata_store import MetadataStore


def initiate_db_connection(host, user_name, database_name, port, password):
    """Establish a database connection

    Returns
    -------
    database_connection
        An established connection to the database

    """

    # attempt to create the database connection
    database_connection = psycopg2.connect("host=%s dbname=%s user=%s password=%s port=%s" %
                                           (host, database_name, user_name, password, port))

    return database_connection


class DbConfigStore(ConfigStore):
    """Database backend for config store
    """

    def __init__(self, host, user, dbname, port, password):
        """Initiate. Setup database connection. Set up common values.
        """
        # establish database connection
        self.database_connection = initiate_db_connection(host, user, dbname, port, password)

        # list of expected port types - these should be specified somewhere, or not presumed at all
        self.port_types = ['inputs', 'outputs', 'parameters']

    # region Model runs
    def read_model_runs(self):
        """Read all model runs

        Returns
        -------
        list
            A list of dicts containing model runs
        """
        # establish a cursor for the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # query database to get model runs
        cursor.execute('SELECT * FROM model_runs;')

        # get result of query
        model_runs = cursor.fetchall()

        return model_runs

    def read_model_run(self, model_run_name):
        """Read a single model run

        Returns
        -------
        dict
            A dictionary containing a model run definition
        """
        # establish a cursor for the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # sql to get model run information
        cursor.execute('SELECT * FROM model_runs WHERE name=%s;', [model_run_name])

        # get result of query
        model_run = cursor.fetchall()

        # check only one model run has been returned
        if len(model_run) > 1:
            # more than one model run returned, a database data error has occurred
            return
        elif len(model_run) == 0:
            # no model run exists with given name
            return

        return model_run[0]

    def write_model_run(self, model_run):
        """Read all systems of systems models

        Argument
        --------
        model_run: dict
            A dictionary containing a model run definition
        """
        # establish a cursor for the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # check model run name is unique
        cursor.execute('SELECT id FROM model_runs WHERE name = %s;', [model_run['name']])

        # get result of query
        model_runs = cursor.fetchall()

        # if one or more model runs are returned, return an error
        if len(model_runs) > 0:
            # model run with same name already exists
            return

        # check given sos model already exists
        cursor.execute('SELECT id FROM sos_models WHERE name=%s;', [model_run['sos_model']])

        # get result of query
        sos_models = cursor.fetchall()

        # check that only one sos model exists with the name given, otherwise return an error
        if len(sos_models) > 1:
            # more than one sos model with the given name - a data error
            return
        elif len(sos_models) == 0:
            # no sos model exists with the name given
            return

        # write model run to database
        cursor.execute('INSERT INTO model_runs (name, sos_model, sos_model_id) VALUES (%s,%s,%s);', [model_run['name'], model_run['sos_model'], sos_models[0]['id']])

        # write data to database
        self.database_connection.commit()

        return

    def update_model_run(self, model_run_name, model_run):
        """Update an existing model run definition

        Argument
        --------
        model_run_name: string
            The name of the model run to update
        model_run: dict
            A dictionary containing a model run definition with those arguments to be updated
        """
        # establish a cursor for the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # check given model run is in the database
        cursor.execute('SELECT id FROM model_runs WHERE name=%s;', [model_run_name])

        # get result of query
        model_runs = cursor.fetchall()

        # if anything but one model run returned, return an error
        if len(model_runs) == 1:

            # if the sos model key has been passed in the definition
            if 'sos_model' in model_run.keys():

                # check the passed model exists
                cursor.execute('SELECT id FROM sos_models WHERE name=%s;', [model_run['sos_model']])

                # get result of query
                sos_model_id = cursor.fetchall()

                # if only one model run id returned, make change
                if len(sos_model_id) == 1:
                    # update table with new sos model
                    cursor.execute('UPDATE model_runs SET sos_model=%s, sos_model_id=%s WHERE name=%s;', [model_run['sos_model'], sos_model_id[0]['id'],model_run_name])

                    # commit update to database
                    self.database_connection.commit()

                else:
                    # return as error as given sos model does not exist
                    return

        else:
            # return an error to the user
            return

        return

    def delete_model_run(self, model_run_name):
        """Delete a model run

        Argument
        --------
        model_run_name: string
            The name of the model run
        """
        # establish a cursor for the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # check given model run is in database
        cursor.execute('SELECT id FROM model_runs WHERE name=%s;', [model_run_name])

        # get result of query
        model_runs = cursor.fetchall()

        # if only one model run returned, delete it
        if len(model_runs) == 1:
            # delete model run
            cursor.execute('DELETE FROM model_runs WHERE name=%s;', [model_run_name])

            # commit the delete action to the database
            self.database_connection.commit()

        else:
            # return an error as no model run, or multiple, returned
            return

        return

    # endregion

    # region System-of-systems models
    def read_sos_models(self):
        """Read all systems of systems models

        Returns
        -------
        list
            A list of dicts containing sos model definitions
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # get sos model details
        cursor.execute('SELECT name FROM sos_models;')

        # get returned data
        sos_models = cursor.fetchall()

        # is some models are returned
        if sos_models is not None:

            # create list to store models
            sos_models_list = []

            # loop through existing models
            for model_name in sos_models:

                # get the model details
                sos_models_list.append(self.read_sos_model(model_name[0]))

        return sos_models_list

    def read_sos_model(self, sos_model_name):
        """Read a single system of systems model

        Argument
        --------
        sos_model_name: string
            The name of the systems of systems model to read

        Returns
        -------
        dict
            The systems of systems model definition
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # get sos model details
        cursor.execute('SELECT * FROM sos_models WHERE name=%s;', [sos_model_name])

        # get returned data
        sos_model = cursor.fetchone()

        if sos_model is not None:
            # get simulation models
            cursor.execute('SELECT * FROM sos_model_simulation_models WHERE sos_model_name=%s;', [sos_model_name])

            # add all returned models to the sector models key in the sos_model definition
            sos_model['sector_models'] = []
            for model in cursor.fetchall():
                sos_model['sector_models'].append(model)

            # get scenario sets
            cursor.execute('SELECT * FROM sos_model_scenarios WHERE sos_model_name=%s;', [sos_model_name])

            # add all returned models to the sector models key in the sos_model definition
            sos_model['scenario_sets'] = []
            for scenario in cursor.fetchall():
                sos_model['scenario_sets'].append(scenario)

            # get sos model dependencies
            cursor.execute('SELECT * FROM sos_model_dependencies WHERE sos_model_name=%s;', [sos_model_name])

            # add all returned dependencies to the dependencies key in the sos_model definition
            sos_model['dependencies'] = []
            for dependency in cursor.fetchall():
                sos_model['dependencies'].append(dependency)

        return sos_model

    def write_sos_model(self, sos_model):
        """Write a systems of systems model

        Argument
        --------
        sos_model: dict
            The definition for a systems os systems model

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # check name does not already exist
        cursor.execute('SELECT name FROM sos_models WHERE name=%s;', [sos_model['name']])

        # get returned data from query
        sos_models = cursor.fetchall()

        # if one of more names are returned
        if len(sos_models) > 0:
            # return an error - sos model already exists, use a different name
            return

        # write sos model, return id
        cursor.execute('INSERT INTO sos_models (name, description) VALUES (%s, %s) RETURNING id;', [sos_model['name'], sos_model['description']])

        # write to database
        self.database_connection.commit()

        # get returned model id
        sos_model_id = cursor.fetchone()

        # write dependencies
        # loop through passed dependencies
        for dependency in sos_model['dependencies']:
            # check dependent models already exist in database - source model
            cursor.execute('SELECT name FROM simulation_models WHERE name=%s;', [dependency['source_model']])

            # get returned data from query
            sos_models = cursor.fetchall()

            # if no names are returned
            if len(sos_models) == 0:
                # return an error - simulation model does not exist
                return

            # check dependent models already exist in database - sink model
            cursor.execute('SELECT name FROM simulation_models WHERE name=%s;', [dependency['sink_model']])

            # get returned data from query
            sos_models = cursor.fetchall()

            # if no names are returned
            if len(sos_models) == 0:
                # return an error - simulation model does not exist
                return

            # sql to write dependency to db
            cursor.execute('INSERT INTO sos_model_dependencies (sos_model_name, source_model, source_output, sink_model, sink_input, lag) VALUES (%s,%s,%s,%s,%s,%s);', [sos_model['name'], dependency['source_model'], dependency['source_model_output'], dependency['sink_model'], dependency['sink_model_input'], dependency['lag']])

            # write to database
            self.database_connection.commit()

        # write sos_model_sim_models
        sim_model_counter = 1
        for sector_model in sos_model['sector_models']:

            # check simulation model already in database
            cursor.execute('SELECT id FROM simulation_models WHERE name=%s;', [sector_model])

            # get returned data from query
            simulation_models = cursor.fetchall()

            # if no names are returned
            if len(simulation_models) == 0:
                # return an error - simulation model does not exist
                return

            # write link between sos model and simulation models
            cursor.execute('INSERT INTO sos_model_simulation_models (sos_model_name, simulation_model_name, simulation_model_id, sos_sim_model_id) VALUES (%s,%s,%s,%s);', [sos_model['name'], sector_model, simulation_models[0]['id'], sim_model_counter])

            # iterate counter
            sim_model_counter += 1

            # write to database
            self.database_connection.commit()

        # write sos_model_scenarios
        for scenario in sos_model['scenario_sets']:
            # check scenario already exists
            cursor.execute('SELECT name FROM scenarios WHERE name=%s;', [scenario])

            # get returned data from query
            scenario_names = cursor.fetchall()

            # if no names are returned
            if len(scenario_names) == 0:
                # return an error - scenario does not exist
                return

            # add data into database
            cursor.execute('INSERT INTO sos_model_scenarios (sos_model_name, scenario_name) VALUES (%s,%s);', [sos_model['name'], scenario])

        # write to database
        self.database_connection.commit()

        return

    def update_sos_model(self, sos_model_name, sos_model):
        """Update a systems of systems model

        Argument
        --------
        sos_model_name: string
            The name of the systems of systems model to update
        sos_model: dict
            The definition of a systems of systems model with only the data to be updated in

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # check sos model exists
        cursor.execute('SELECT id FROM sos_models WHERE name=%s;', [sos_model_name])

        # get result of query
        sos_models = cursor.fetchall()

        # if no model with the name exists, return
        if len(sos_models) == 0:
            # no model with the name given
            return

        # update dependencies if passed
        if 'dependencies' in sos_model.keys():

            # loop through dependencies
            for dependency in sos_model['dependencies']:
                # update details

                if 'source_model' in dependency.keys():
                    # check dependent models already exist in database - source model
                    cursor.execute('SELECT name FROM simulation_models WHERE name=%s;', [dependency['source_model']])

                    # get returned data from query
                    sos_models = cursor.fetchall()

                    # if no names are returned
                    if len(sos_models) == 0:
                        # return an error - simulation model does not exist
                        return

                    # update
                    cursor.execute('UPDATE dependencies SET source_model=%s WHERE name=%s;' [dependency['source_model'], sos_model_name])

                    # write to database
                    self.database_connection.commit()

                if 'sink_model' in dependency.keys():
                    # check dependent models already exist in database - sink model
                    cursor.execute('SELECT name FROM simulation_models WHERE name=%s;', [dependency['sink_model']])

                    # get returned data from query
                    sos_models = cursor.fetchall()

                    # if no names are returned
                    if len(sos_models) == 0:
                        # return an error - simulation model does not exist
                        return

                    # update
                    cursor.execute('UPDATE dependencies SET sink_model=%s WHERE name=%s;', [dependency['source_model'], sos_model_name])

                    # write to database
                    self.database_connection.commit()

                if 'source_model_output' in dependency.keys():
                    # update
                    cursor.execute('UPDATE dependencies SET source_model_output=%s WHERE name=%s;', [dependency['source_model_output'], sos_model['name']])

                    # write to database
                    self.database_connection.commit()

                if 'sink_model_input' in dependency.keys():
                    # update
                    cursor.execute('UPDATE dependencies SET sink_model_input=%s WHERE name=%s;', [dependency['sink_model_input'], sos_model_name])

                    # write to database
                    self.database_connection.commit()

                if 'lag' in dependency.keys():
                    # update
                    cursor.execute('UPDATE dependencies SET lag=%s WHERE name=%s;', [dependency['lag'], sos_model_name])

                    # write to database
                    self.database_connection.commit()

        # need to update methods as does not all for multiple as different rows
        # update sos_model_simulation_models if passed
        if 'sector_models' in sos_model.keys():

            # loop through the sector models
            for sector_model in sos_model['sector_models']:

                # check simulation model already in database
                cursor.execute('SELECT name FROM simulation_models WHERE name=%s;', [sector_model])

                # get returned data from query
                sos_models = cursor.fetchall()

                # if no names are returned
                if len(sos_models) == 0:
                    # return an error - simulation model does not exist
                    return

            # if all simulation models in database, attempt to update the sos model sim model relation

            # get what is already in the database
            cursor.execute('SELECT *  FROM sos_model_simulation_models WHERE sos_model_name=%s;', [sos_model_name])

            # get result from query
            sos_model_sim_models = cursor.fetchall()

            for sector_model in sos_model['sector_models']:
                # already checked if sim model in database
                # start by seeing if in relation already - loop through existing data
                for sim_model in sos_model_sim_models:
                    if sim_model['simulation_model_name'] == sector_model:
                        # return an error as already in relation so don't need to add again
                        return

                # if here, not in relation so add sim model to sos model sim model relation
                # get max id count in relation
                cursor.execute('SELECT max(sos_sim_model_id) FROM sos_model_simulation_models WHERE sos_model_name=%s;', [sos_model_name])

                # get result of query
                sos_sim_model_id = cursor.fetchone()

                # write sim model to sos model sim models relation
                cursor.execute('INSERT INTO sos_model_simulation_models (sos_model_name, simulation_model_name, sos_sim_model_id) VALUES (%s,%s,%s);', [sos_model_name, sector_model, sos_sim_model_id[0]])

                # write update to database
                self.database_connection.commit()

        # need to update methods as does not allow multiple as different rows
        # update sos_model_scenarios if passed
        if 'scenario_sets' in sos_model.keys():

            # loop through passed scenarios
            for scenario in sos_model['scenario_sets']:
                # check scenario already exists
                cursor.execute('SELECT name FROM scenarios WHERE name=%s;', [scenario])

                # get returned data from query
                scenario_names = cursor.fetchall()

                # if no names are returned
                if len(scenario_names) == 0:
                    # return an error - scenario does not exist
                    return

                # add data into database
                cursor.execute('UPDATE sos_model_scenarios SET scenario_name=%s WHERE sos_model_name=%s;', [scenario, sos_model_name])

            # write to database
            self.database_connection.commit()

        # update sos model description if passed
        if 'description' in sos_model.keys():

            # sql to update description
            cursor.execute('UPDATE sos_models SET description=%s WHERE name=%s;', [sos_model['description'], sos_model_name])

            # write to database
            self.database_connection.commit()

        return

    def delete_sos_model(self, sos_model_name):
        """Delete a systems of systems model

        Argument
        --------
        sos_model_name: string
            The name of the systems of systems model to delete

        """

        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # check sos model exists
        cursor.execute('SELECT id FROM sos_models WHERE name=%s;', [sos_model_name])

        # get result of query
        sos_models = cursor.fetchall()

        # if none returned, exit with error
        if len(sos_models) == 0:
            # no model with the given name found
            return

        # delete from sos_model_simulation_models
        cursor.execute('DELETE FROM sos_model_simulation_models WHERE sos_model_name = %s;', [sos_model_name])

        # run query to delete from db
        self.database_connection.commit()

        # delete from sos_model_scenarios
        cursor.execute('DELETE FROM sos_model_scenarios WHERE sos_model_name = %s;', [sos_model_name])

        # run query to delete from db
        self.database_connection.commit()

        # delete from sos_model_dependencies
        cursor.execute('DELETE FROM sos_model_dependencies WHERE sos_model_name = %s;', [sos_model_name])

        # run query to delete from db
        self.database_connection.commit()

        # delete from model_run
        cursor.execute('DELETE FROM model_runs WHERE sos_model = %s;', [sos_model_name])

        # run query to delete from db
        self.database_connection.commit()

        # delete from sos_models
        cursor.execute('DELETE FROM sos_models WHERE name = %s;', [sos_model_name])

        # run query to delete from db
        self.database_connection.commit()

        return
    # endregion

    # region Models
    def read_models(self):
        """Read all simulation models

        Returns
        -------
        list
            A list of dictionaries for the models returned
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # run sql call
        cursor.execute('SELECT * FROM simulation_models')

        # get returned data
        simulation_models = cursor.fetchall()

        # loop through the returned simulation models to get their details
        # create dictionary to store simulation models
        simulation_model_list = {}
        # loop through known models
        for simulation_model in simulation_models:

            # get details of the models from the read call and add to list
            simulation_model_list[simulation_model['name']] = self.read_model(simulation_model['name'])

        # return data to user
        return simulation_model_list

    def read_model(self, model_name):
        """Read a simulation model

        Argument
        --------
        model_name: string
            The name of the model to read

        Returns
        -------
        dict
            A dictionary of the model definition returned
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # run sql call
        cursor.execute('SELECT * FROM simulation_models WHERE name=%s', [model_name])

        # get returned data
        simulation_model = cursor.fetchall()

        # check if a simulation model has been found and only proceed if so
        if len(simulation_model) == 1:
            # get port types for model - inputs, outputs, parameters
            # run sql call
            for port_type in self.port_types:

                # get specification details by joining sim model port and specification relation using the specification id (unique to each specification)
                cursor.execute('SELECT s.* FROM simulation_model_port smp JOIN specifications s ON smp.specification_id=s.id WHERE smp.model_name=%s AND port_type=%s;', [model_name, port_type])

                # get returned data
                port_data = cursor.fetchall()

                # add port data to model dictionary
                simulation_model[0][port_type] = port_data

        # return data to user
        return simulation_model

    def write_model(self, model):
        """Write a simulation model to the database

        Argument
        --------
        model: dictionary
            A model definition

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # write model
        # write the model to the database
        cursor.execute('INSERT INTO simulation_models (name, description, interventions, wrapper_location) VALUES (%s,%s,%s,%s) RETURNING id;', [model['name'], model['description'], json.dumps(model['interventions']), model['wrapper_location']])

        # commit changes to database
        self.database_connection.commit()

        # get returned data
        simulation_model_id = cursor.fetchone()

        # write the specification to the database (metadata on the inputs/outputs/parameters)

        # loop through port types
        for port_type in self.port_types:
            # if port type is a key in given model definition, add all specifications to database
            if port_type in model.keys():
                # loop through the specifications for the port type
                for spec in model[port_type]:
                    # should probably check it is already not in the database first

                    # add specification to database
                    cursor.execute('INSERT INTO specifications (name, description, dimensions, unit, suggested_range, absolute_range) VALUES (%s,%s,%s,%s,%s,%s) RETURNING id;', [spec['name'], spec['description'], spec['dimensions'], spec['unit'], spec['suggested_range'], spec['absolute_range']])

                    # commit changes to database
                    self.database_connection.commit()

                    # get returned data
                    specification_id = cursor.fetchone()

                    # write to port
                    # write model and specification to port table, including the specification id
                    cursor.execute('INSERT INTO simulation_model_port (model_name, model_id, specification_name, specification_id, port_type) VALUES (%s, %s, %s, %s, %s);', [model['name'], simulation_model_id[0], spec['name'], specification_id[0], port_type])

                    # commit changes to database
                    self.database_connection.commit()

        # return data to user
        return

    def update_model(self, model_name, model):
        """Update a simulation model

        Argument
        --------
        model_name: string
            The name of the model to update
        model:
             A model definition with only the fields to be updated

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # update the model description is passed
        if 'description' in model.keys():
            cursor.execute('UPDATE simulation_models SET description=%s WHERE name=%s;', [model['description'], model_name])

        # update the intervention list is passed
        if 'interventions' in model.keys():
            cursor.execute('UPDATE simulation_models SET interventions=%s WHERE name=%s;', [json.dumps(model['interventions']), model_name])

        # update the wrapper location if passed
        if 'wrapper_location' in model.keys():
            cursor.execute('UPDATE simulation_models SET wrapper_location=%s WHERE name=%s;', [model['wrapper_location'], model_name])

        # commit changes to database
        self.database_connection.commit()

        # update any of the port types if passed
        # need to figure out how to update an inputs/output/parameter and the specification
        for port_type in self.port_types:

            # if the port type has been passed to be updated
            if port_type in model.keys():

                # loop through each specification for the port type
                for spec in model[port_type]:

                    # get the id of the specification to update - based on name, model and port
                    cursor.execute('SELECT specification_id FROM simulation_model_port WHERE port_type=%s and specification_name=%s and model_name=%s;', [port_type, spec['name'], model_name])

                    # get result of query
                    spec_id = cursor.fetchall()

                    # check for possible errors related to the existence of the specification
                    if len(spec_id) > 1:
                        # return an error, more than one specification id returned
                        return
                    elif len(spec_id) == 0:
                        # return an error - no matching specification found to be updated
                        return

                    # check for each key in the specification and update if present
                    if 'name' in spec.keys():
                        continue
                    if 'description' in spec.keys():
                        # run update sql
                        cursor.execute('UPDATE specifications SET description = %s WHERE id=%s', [spec['description'], spec_id])
                    if 'dimensions' in spec.keys():
                        # run update sql
                        cursor.execute('UPDATE specifications SET dimensions = %s WHERE id=%s',                                       [spec['dimensions'], spec_id])
                    if 'unit' in spec.keys():
                        # run update sql
                        cursor.execute('UPDATE specifications SET unit = %s WHERE id=%s', [spec['unit'], spec_id])
                    if 'suggested_range' in spec.keys():
                        # run update sql
                        cursor.execute('UPDATE specifications SET suggested_range = %s WHERE id=%s',[spec['suggested_range'], spec_id])
                    if 'absolute_range' in spec.keys():
                        # run update sql
                        cursor.execute('UPDATE specifications SET absolute_range = %s WHERE id=%s',[spec['absolute_range'], spec_id])

                    # commit changes to database
                    self.database_connection.commit()

        return

    def delete_model(self, model_name):
        """Delete a simulation model

        Argument
        --------
        model_name: string
            The name of the model to be deleted

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # delete specifications if unique to model
        # get a list of the specifications linked to the model name
        cursor.execute('SELECT * FROM simulation_model_port WHERE model_name=%s;', [model_name])

        # loop through returned model specifications
        for specification in cursor.fetchall():

            # check if the specification is used by any other models
            cursor.execute('SELECT COUNT(*) FROM simulation_model_port WHERE specification_id=%s;', [specification['id']])

            # get the count from the query
            specification_count = cursor.fetchone()

            # if the count is only 1, safe to delete specification, otherwise leave it
            if specification_count == 1:
                cursor.execute('DELETE FROM specifications WHERE id=%s;', [specification['id']])

                # commit changes to database
                self.database_connection.commit()

        # run sql call to delete from simulation model port
        cursor.execute('DELETE FROM simulation_model_port WHERE model_name=%s;', [model_name])

        # commit changes to database
        self.database_connection.commit()

        # run sql call to delete from simulation models
        cursor.execute('DELETE FROM simulation_models WHERE name=%s;', [model_name])

        # commit changes to database
        self.database_connection.commit()

        # get the number of rows deleted and return
        affected_rows = cursor.rowcount

        return
    # endregion

    # region Scenarios
    def read_scenarios(self):
        """Read list of scenarios

         Returns
        -------
        list
            A list of scenarios as dictionary's
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # run sql call
        cursor.execute('SELECT * FROM scenarios')

        # get returned data
        scenarios = cursor.fetchall()

        # return data to user
        return scenarios

    def read_scenario(self, scenario_name):
        """Read a scenario

        Argument
        --------
        scenario_name: string
            The name of the scenario to read

        Returns
        -------
        dict
            A scenario definition
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # run sql call
        cursor.execute('SELECT * FROM scenarios WHERE name=%s', [scenario_name])

        # get returned data
        scenario = cursor.fetchone()

        # return data to user
        return scenario

    def write_scenario(self, scenario):
        """Write a scenario

        Argument
        --------
        scenario: dict
            A scenario definition

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # run sql call
        sql = 'INSERT INTO scenarios (name, description) VALUES (%s,%s) RETURNING id;'
        cursor.execute(sql, [scenario['name'], scenario['description']])

        # commit changes to database
        self.database_connection.commit()

        # get id of new scenario - checks it has been written in
        scenario_id = cursor.fetchone()

        return

    def update_scenario(self, scenario_name, scenario):
        """Update a scenario

        Argument
        --------
        scenario_name: string
            The name of the scenario to update
        scenario: dict
            The project configuration

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # run sql call
        cursor.execute('UPDATE scenarios SET description = %s WHERE name=%s', [scenario['description'], scenario_name])

        # commit changes to database
        self.database_connection.commit()

        return

    def delete_scenario(self, scenario_name):
        """Delete a scenario

        Argument
        --------
        scenario_name: string
            The name of the scenario to delete
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # check passed scenario exists
        cursor.execute('SELECT id FROM scenarios WHERE name=%s;', [scenario_name])

        # get returned data - scenario id
        scenario_id = cursor.fetchall()

        if len(scenario_id) > 0:

            # need to check for and delete any scenario variants that are associated with this scenario
            cursor.execute('DELETE FROM scenario_variants WHERE scenario_name=%s;', [scenario_name])

            # commit changes to database
            self.database_connection.commit()

            # run sql call
            cursor.execute('DELETE FROM scenarios WHERE name=%s', [scenario_name])

            # commit changes to database
            self.database_connection.commit()

        return
    # endregion

    # region Scenario Variants
    def read_scenario_variants(self, scenario_name):
        """Read scenario variants

        Argument
        --------
        scenario_name: string
            The name of the scenario name to return variants for

        Returns
        -------
        dict
            A list scenario variants as dictionaries
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # run sql call
        cursor.execute('SELECT sv.*, v.description, v.data FROM scenario_variants sv JOIN variants v ON sv.variant_name = v.name WHERE sv.scenario_name=%s', [scenario_name])

        # get returned data
        scenario_variants = cursor.fetchall()

        # return data to user
        return scenario_variants

    def read_scenario_variant(self, scenario_name, variant_name):
        """Read scenario variants

        Argument
        --------
        scenario_name: string
            The name of the scenario
        variant_name: string
            The name of scenario variant to return

        Returns
        -------
        dict
            A scenario variant definition
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        # run sql call
        cursor.execute('SELECT sv.*, v.description, v.data FROM scenario_variants sv JOIN variants v ON sv.variant_name = v.name WHERE sv.scenario_name=%s AND sv.variant_name=%s', [scenario_name, variant_name])

        # get returned data
        scenario_variant = cursor.fetchall()

        # return data to user
        return scenario_variant

    def write_scenario_variant(self, scenario_name, variant):
        """Write scenario variant

        Argument
        --------
        scenario_name: string
            The name of the scenario the variant is associated with
        variant: dict
            The variant definition

        Returns
        -------
        integer
            The id of the added scenario variant
        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # here is a dependency on the given scenario name already existing
        # - this can be enforced in the database by a foreign key, but should probably check here
        # - would also enable access to the scenario id
        cursor.execute('SELECT id FROM scenarios WHERE name=%s;', [scenario_name])

        # get returned data
        scenario_id = cursor.fetchall()

        # if a scenario_id has been found
        if len(scenario_id) == 1:

            # run sql to add data to variants to database
            cursor.execute('INSERT INTO variants (name, description, data) VALUES (%s,%s,%s) RETURNING id;', [variant['variant_name'], variant['description'], json.dumps(variant['data'])])

            # commit changes to database
            self.database_connection.commit()

            # get returned data
            variant_id = cursor.fetchone()

            # run sql call
            sql = 'INSERT INTO scenario_variants (scenario_name, variant_name, scenario_id) VALUES (%s,%s,%s) RETURNING id;'
            cursor.execute(sql, [variant['scenario_name'], variant['variant_name'], scenario_id[0]['id']])

            # commit changes to database
            self.database_connection.commit()

            # get returned data
            scenario_variant_id = cursor.fetchone()

            return scenario_variant_id
        else:
            return

    def update_scenario_variant(self, scenario_name, variant_name, variant):
        """Update scenario variant

        Argument
        --------
        scenario_name: string
            The name of the scenario the variant is associated with
        variant_name: string
            The name of variant to be updated
        variant: dict
            The variant definition containing the data to be updated

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.RealDictCursor)

        if 'description' in variant.keys() and 'data' in variant.keys():

            # run sql call
            cursor.execute('UPDATE variants SET description=%s AND data=%s WHERE name = %s RETURNING id;', [variant['description'], json.dumps(variant['data']), variant_name])

        elif 'description' in variant.keys() and 'data' not in variant.keys():

            # run sql call
            cursor.execute('UPDATE variants SET description=%s WHERE name = %s RETURNING id;', [variant['description'], variant_name])

        elif 'description' not in variant.keys() and 'data' in variant.keys():

            # run sql call
            cursor.execute('UPDATE variants SET data=%s WHERE name = %s RETURNING id;', [json.dumps(variant['data']), variant_name])

        # commit changes to database
        self.database_connection.commit()

        # get the number of rows deleted and return
        affected_rows = cursor.rowcount

        return affected_rows

    def delete_scenario_variant(self, scenario_name, variant_name):
        """"Delete scenario variant

        Argument
        --------
        scenario_name: string
            The name of the scenario the variant to be deleted is associated with
        variant_name: string
            The name of the variant to be deleted

        Returns
        -------
        integer
            The number of rows deleted

        """
        # establish a cursor to read the database
        cursor = self.database_connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

        # run sql call
        sql = 'DELETE FROM scenario_variants WHERE scenario_name=%s AND variant_name=%s'
        cursor.execute(sql, [scenario_name, variant_name])

        # commit changes to database
        self.database_connection.commit()

        # run sql call
        cursor.execute('DELETE FROM variants WHERE name=%s', [scenario_name, variant_name])

        # commit changes to database
        self.database_connection.commit()

        # get the number of rows deleted and return
        affected_rows = cursor.rowcount

        return affected_rows

    # endregion

    # region Narratives
    def read_narrative(self, sos_model_name, narrative_name):
        raise NotImplementedError()
    # endregion

    # region Strategies
    def read_strategies(self, model_run_name):
        raise NotImplementedError()

    def write_strategies(self, model_run_name, strategies):
        raise NotImplementedError()
    # endregion


class DbMetadataStore(MetadataStore):
    """Database backend for metadata store
    """
    # region Units
    def read_unit_definitions(self):
        raise NotImplementedError()
    # endregion

    # region Dimensions
    def read_dimensions(self):
        raise NotImplementedError()

    def read_dimension(self, dimension_name):
        raise NotImplementedError()

    def write_dimension(self, dimension):
        raise NotImplementedError()

    def update_dimension(self, dimension_name, dimension):
        raise NotImplementedError()

    def delete_dimension(self, dimension_name):
        raise NotImplementedError()
    # endregion


class DbDataStore(DataStore):
    """Database backend for data store
    """
    # region Scenario Variant Data
    def read_scenario_variant_data(self, scenario_name, variant_name, variable, timestep=None):
        raise NotImplementedError()

    def write_scenario_variant_data(self, scenario_name, variant_name,
                                    data_array, timestep=None):
        raise NotImplementedError()
    # endregion

    # region Narrative Data
    def read_narrative_variant_data(self, narrative_name, variant_name, variable,
                                    timestep=None):
        raise NotImplementedError()

    def write_narrative_variant_data(self, narrative_name, variant_name, data_array,
                                     timestep=None):
        raise NotImplementedError()

    def read_model_parameter_default(self, model_name, parameter_name):
        raise NotImplementedError()

    def write_model_parameter_default(self, model_name, parameter_name, data):
        raise NotImplementedError()
    # endregion

    # region Interventions
    def read_interventions(self, sector_model_name):
        raise NotImplementedError()

    def read_initial_conditions(self, sector_model_name):
        raise NotImplementedError()
    # endregion

    # region State
    def read_state(self, modelrun_name, timestep, decision_iteration=None):
        raise NotImplementedError()

    def write_state(self, state, modelrun_name, timestep, decision_iteration=None):
        raise NotImplementedError()
    # endregion

    # region Conversion coefficients
    def read_coefficients(self, source_dim, destination_dim):
        raise NotImplementedError

    def write_coefficients(self, source_dim, destination_dim, data):
        raise NotImplementedError()
    # endregion

    # region Results
    def read_results(self, modelrun_name, model_name, output_spec, timestep=None,
                     decision_iteration=None):
        raise NotImplementedError()

    def write_results(self, data_array, modelrun_name, model_name, timestep=None,
                      decision_iteration=None):
        raise NotImplementedError()

    def prepare_warm_start(self, modelrun_id):
        raise NotImplementedError()
    # endregion



"""Set up a smif database
"""
import subprocess
from os import environ, listdir
from os.path import dirname, isfile, join, realpath


def run_migrations(migrations_dir, migration_direction, database_name):
    """Run specified database migrations
    """

    # run migrations given parameter passed
    if migration_direction == 'up':
        ran = up_migrations(migrations_dir, database_name)
    elif migration_direction == 'down':
        ran = down_migrations(migrations_dir, database_name)

    return ran


def down_migrations(working_path, database_name):
    """Run down migrations to resolve database to blank
    """

    # run key file to remove foreign keys from database
    file = 'remove-foreign_keys.sql'
    # check file exists
    if isfile(join(working_path, file)):
        # run sql file silently
        subprocess.run(['psql', '-d', database_name, '-q', '-f', join(working_path, file)])
    else:
        return False

    # loop through files to remove tables
    for file in listdir(working_path):

        # check if file is a down migration
        if file[0:4] == 'down':
            # run sql file silently
            subprocess.run(['psql', '-d', database_name, '-q', '-f', join(working_path, file)])

    return True


def up_migrations(working_path, database_name):
    """Run up migrations to build database
    """

    # declare name of foreign key file
    # run key file to add foreign keys between relations
    fky_file = 'add-foreign_keys.sql'

    # check file exists - if not stop now
    if isfile(join(working_path, fky_file)) is False:
        return False

    # loop through files
    for file in listdir(working_path):

        # check if file is a down migration
        if file[0:2] == 'up':
            # run sql file silently
            subprocess.run(['psql', '-d', database_name, '-q', '-f', join(working_path, file)])

    # run key file to add foreign keys between relations
    # run sql file silently
    subprocess.run(['psql', '-d', database_name, '-q', '-f', join(working_path, fky_file)])

    return True


def main(database_name):
    """Run database setup

    Relies on standard Postgres environment variables being set to configure the database
    connection: https://www.postgresql.org/docs/current/libpq-envars.html
    """
    # get the base directory and build the path for the database config files
    migrations_dir = join(dirname(realpath(__file__)), 'migrations')

    # run migrations - delete existing relations than add new set
    # run the migrations - down first in case anything exists already
    # run migrations - down
    run_migrations(migrations_dir, 'down', database_name)

    # run the migrations - up to build the database
    # run migrations - up
    run_migrations(migrations_dir, 'up', database_name)

    # confirm migrations have been run
    print('Finished building database')


if __name__ == '__main__':
    # run main when file called
    try:
        DB_NAME = environ['PGDATABASE']
    except KeyError:
        print('PGDATABASE environment variable must be set to configure database name')
        exit(1)
    main(DB_NAME)



"""Test Coordinates metadata
"""
from collections import OrderedDict

from pytest import mark, raises
from smif.metadata import Coordinates


class CustomMapping():
    """Custom mapping that doesn't inherit dict
    """
    def __init__(self, key_values):
        self._data = {}
        for key, value in key_values:
            self._data[key] = value

    def __getitem__(self, key):
        return self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)


class TestCoordinates():
    """Coordinates has name and elements

    Equivalent to an Element Set (OpenMI)
    Can be passed as coords to DataSpec or DataArray
    """
    def test_construct_with_element_names(self):
        """Create a Coordinates with name and elements (list of ids)
        """
        name = 'building_categories'
        element_ids = ['residential', 'commercial', 'industrial']

        building_sectors = Coordinates(name, element_ids)

        assert building_sectors.name == name
        assert building_sectors.ids == element_ids
        assert building_sectors.names == element_ids
        assert building_sectors.elements == [
            {'name': 'residential'},
            {'name': 'commercial'},
            {'name': 'industrial'}
        ]

    @mark.parametrize("elements", [
        [
            {'name': 'residential', 'sector': True},
            {'name': 'commercial', 'sector': True},
            {'name': 'industrial', 'sector': True},
        ],
        [
            OrderedDict([('name', 'residential'), ('sector', True)]),
            OrderedDict([('name', 'commercial'), ('sector', True)]),
            OrderedDict([('name', 'industrial'), ('sector', True)]),
        ],
        [
            CustomMapping([('name', 'residential'), ('sector', True)]),
            CustomMapping([('name', 'commercial'), ('sector', True)]),
            CustomMapping([('name', 'industrial'), ('sector', True)]),
        ]
    ])
    def test_construct_with_elements(self, elements):
        """Create a Coordinates with name and elements (list of dicts)
        """
        name = 'building_categories'

        building_sectors = Coordinates(name, elements)

        assert building_sectors.name == name
        assert building_sectors.ids == ['residential', 'commercial', 'industrial']
        assert building_sectors.names == ['residential', 'commercial', 'industrial']
        assert building_sectors.elements == elements

    def test_name_dim_alias(self):
        """Name and dim refer to the same label
        """
        coords = Coordinates('name', ['test'])
        assert coords.name == 'name'
        assert coords.dim == 'name'

        coords.name = 'name2'
        assert coords.name == 'name2'
        assert coords.dim == 'name2'

        coords.dim = 'name3'
        assert coords.name == 'name3'
        assert coords.dim == 'name3'

    def test_coordinates_must_have_elements(self):
        """A Coordinates must have one or more elements
        """
        with raises(ValueError) as ex:
            Coordinates('zero_d', [])
        assert "must not be empty" in str(ex)

    def test_elements_must_have_name(self):
        """Coordinates elements must have "name"
        """
        elements = [
            {"description": "Petrol", "state": "liquid"},
            {"description": "Diesel", "state": "liquid"},
            {"description": "Coal", "state": "solid"},
        ]
        with raises(KeyError) as ex:
            Coordinates('fossil_fuels', elements)
        assert "must have a name" in str(ex)

    def test_elements_must_be_finite(self):
        """Only accept finite Coordinatess
        """
        def natural_numbers():
            i = 0
            while True:
                yield i
                i += 1

        elements = natural_numbers()

        with raises(ValueError) as ex:
            Coordinates('natural_numbers', elements)
        assert "must be finite" in str(ex)

    def test_eq(self):
        """Equality based on equivalent name and elements
        """
        a = Coordinates('name', [1, 2, 3])
        b = Coordinates('name', [
            {'name': 1},
            {'name': 2},
            {'name': 3}
        ])
        c = Coordinates('another', [1, 2, 3])
        d = Coordinates('name', [2, 3, 4])
        e = Coordinates('name', [
            {'name': 1, 'note': 'meta'},
            {'name': 2, 'note': 'meta'},
            {'name': 3, 'note': 'meta'}
        ])
        assert a == b
        assert a != c
        assert a != d
        assert a != e



"""Test command line interface
"""

import os
import subprocess
import sys
from tempfile import TemporaryDirectory
from unittest.mock import call, patch

import smif
from pytest import fixture
from smif.cli import confirm, parse_arguments, setup_project_folder


@fixture
def tmp_sample_project(tmpdir_factory):
    test_folder = tmpdir_factory.mktemp("smif")
    subprocess.run(
        ["smif", "-v", "setup", "-d", str(test_folder)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    return str(test_folder)


def get_args(args):
    """Get args object from list of strings
    """
    parser = parse_arguments()
    return parser.parse_args(args)


def test_parse_arguments():
    """Setup a project folder argument parsing
    """
    with TemporaryDirectory() as project_folder:
        args = get_args(['setup', '-d', project_folder])

        expected = project_folder
        actual = args.directory
        assert actual == expected

        # Ensure that the `setup_project_folder` function is called when `setup`
        # command is passed to the cli
        assert args.func.__name__ == 'setup_project_folder'


def test_fixture_single_run(tmp_sample_project):
    """Test running the (default) binary-filesystem-based single_run fixture
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "-v", "run", "-d", config_dir,
                             "energy_central"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print(output.stdout.decode("utf-8"))
    print(output.stderr.decode("utf-8"), file=sys.stderr)
    assert "Running energy_central" in str(output.stderr)
    assert "Model run 'energy_central' complete" in str(output.stdout)


def test_fixture_single_run_csv(tmp_sample_project):
    """Test running the csv-filesystem-based single_run fixture
    """
    output = subprocess.run(
        ["smif", "-v", "run", "-i", "local_csv", "-d", tmp_sample_project,
         "energy_central"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    print(output.stdout.decode("utf-8"))
    print(output.stderr.decode("utf-8"), file=sys.stderr)
    assert "Running energy_central" in str(output.stderr)
    assert "Model run 'energy_central' complete" in str(output.stdout)


def test_fixture_single_run_warm(tmp_sample_project):
    """Test running the (default) single_run fixture with warm setting enabled
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "-v", "run", "-w", "-d", config_dir,
                             "energy_central"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print(output.stdout.decode("utf-8"))
    print(output.stderr.decode("utf-8"), file=sys.stderr)
    assert "Running energy_central" in str(output.stderr)
    assert "Model run 'energy_central' complete" in str(output.stdout)


def test_fixture_batch_run(tmp_sample_project):
    """Test running the multiple modelruns using the batch_run option
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "-v", "run", "-b", "-d", config_dir,
                             os.path.join(config_dir, "batchfile")],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print(output.stdout.decode("utf-8"))
    print(output.stderr.decode("utf-8"), file=sys.stderr)
    assert "Running energy_water_cp_cr" in str(output.stderr)
    assert "Model run 'energy_water_cp_cr' complete" in str(output.stdout)
    assert "Running energy_central" in str(output.stderr)
    assert "Model run 'energy_central' complete" in str(output.stdout)


def test_fixture_list_runs(tmp_sample_project):
    """Test running the filesystem-based single_run fixture
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "list", "-d", config_dir], stdout=subprocess.PIPE)
    assert "energy_water_cp_cr" in str(output.stdout)
    assert "energy_central" in str(output.stdout)


def test_setup_project_folder():
    """Test contents of the setup project folder
    """
    with TemporaryDirectory() as project_folder:
        args = get_args(['setup', '-d', project_folder])
        setup_project_folder(args)

        assert os.path.exists(project_folder)

        folder_list = ['config', 'data', 'models', 'planning']
        for folder in folder_list:
            folder_path = os.path.join(project_folder, folder)

            assert os.path.exists(folder_path)


@patch('builtins.input', return_value='y')
def test_confirm_yes(input):
    assert confirm()


@patch('builtins.input', return_value='n')
def test_confirm_no(input):
    assert not confirm()


@patch('builtins.input', return_value='')
def test_confirm_default_response(input):
    assert not confirm()


@patch('builtins.input', return_value='n')
@patch('builtins.print')
def test_confirm_default_message(mock_print, input):
    confirm()
    input.assert_has_calls([call('Confirm [n]|y: ')])


@patch('builtins.input', return_value='n')
@patch('builtins.print')
def test_confirm_custom_message(mock_print, input):
    confirm('Create directory?', True)
    input.assert_has_calls([call('Create directory? [y]|n: ')])


@patch('builtins.input', side_effect=['invalid', 'y'])
@patch('builtins.print')
def test_confirm_repeat_message(mock_print, input):
    confirm()
    input.assert_has_calls([call('Confirm [n]|y: '), call('Confirm [n]|y: ')])
    mock_print.assert_called_with('please enter y or n.')


def test_version_display():
    """Expect version number from `smif -V`
    """
    output = subprocess.run(['smif', '-V'], stdout=subprocess.PIPE)
    assert smif.__version__ in str(output.stdout)


def test_verbose_debug():
    """Expect debug message from `smif -vv`
    """
    output = subprocess.run(['smif', '-vv'], stderr=subprocess.PIPE)
    assert 'DEBUG' in str(output.stderr)


def test_verbose_debug_alt():
    """Expect debug message from `smif --verbose --verbose`
    """
    output = subprocess.run(['smif', '--verbose', '--verbose'], stderr=subprocess.PIPE)
    assert 'DEBUG' in str(output.stderr)



"""Test relative timesteps
"""
from pytest import raises
from smif.exception import SmifTimestepResolutionError
from smif.metadata import RelativeTimestep


class TestRelativeTimestep(object):
    def test_create_from_name(self):
        """Should be able to create enum from string parameter
        """
        assert RelativeTimestep.ALL == RelativeTimestep.from_name("ALL")
        assert RelativeTimestep.BASE == RelativeTimestep.from_name("BASE")
        assert RelativeTimestep.CURRENT == RelativeTimestep.from_name("CURRENT")
        assert RelativeTimestep.PREVIOUS == RelativeTimestep.from_name("PREVIOUS")
        with raises(ValueError) as ex:
            RelativeTimestep.from_name("UNKNOWN")
        assert "Relative timestep 'UNKNOWN' is not recognised" in str(ex)

    def test_resolve(self):
        """Should be able to resolve current/previous/base timesteps
        """
        timesteps = list(range(5))
        rel = RelativeTimestep.CURRENT
        assert rel.resolve_relative_to(3, timesteps) == 3
        rel = RelativeTimestep.PREVIOUS
        assert rel.resolve_relative_to(3, timesteps) == 2
        rel = RelativeTimestep.BASE
        assert rel.resolve_relative_to(3, timesteps) == 0
        rel = RelativeTimestep.ALL
        assert rel.resolve_relative_to(3, timesteps) is None

        with raises(SmifTimestepResolutionError) as ex:
            rel.resolve_relative_to(99, timesteps)
        assert "is not present" in str(ex)

        rel = RelativeTimestep.PREVIOUS
        with raises(SmifTimestepResolutionError) as ex:
            rel.resolve_relative_to(0, timesteps)
        assert "has no previous timestep" in str(ex)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""smif
"""
from __future__ import division, print_function, absolute_import

import pkg_resources
import warnings

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


try:
    __version__ = pkg_resources.get_distribution(__name__).version
except Exception:
    __version__ = 'unknown'

# Filter out warnings arising from some installed combinations of scipy/numpy
# - problem and fix discussed in [numpy/numpy#432](https://github.com/numpy/numpy/pull/432)
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")



"""Test ModelData
"""
# pylint: disable=redefined-outer-name
from copy import copy
from unittest.mock import Mock

import numpy as np
from pytest import fixture, raises
from smif.data_layer import DataHandle
from smif.data_layer.data_array import DataArray
from smif.data_layer.data_handle import ResultsHandle
from smif.exception import (SmifDataError, SmifDataMismatchError,
                            SmifDataNotFoundError, SmifTimestepResolutionError)
from smif.metadata import Spec
from smif.model import SectorModel, SosModel


@fixture(scope='function')
def mock_store(sample_dimensions, annual, get_sector_model, empty_store):
    """Store with minimal setup
    """
    store = empty_store

    store.write_unit_definitions([
        'people',
        'million people = 1000000 * people'
    ])

    for dim in sample_dimensions:
        store.write_dimension(dim)

    # energy demand model
    store.write_model(get_sector_model)
    for param in get_sector_model['parameters']:
        spec = Spec.from_dict(param)
        data = DataArray(spec, np.array(42))
        store.write_model_parameter_default('energy_demand', spec.name, data)

    store.write_initial_conditions('energy_demand', [])
    data = {
        'water_asset_a': {
            'name': 'water_asset_a',
            'build_year': 2010,
            'capacity': 50,
            'location': None,
            'sector': ''
        },
        'water_asset_b': {
            'name': 'water_asset_b',
            'build_year': 2015,
            'capacity': 150,
            'location': None,
            'sector': ''
        },
        'water_asset_c': {
            'name': 'water_asset_c',
            'capacity': 100,
            'build_year': 2015,
            'location': None,
            'sector': ''
        }
    }
    store.write_interventions('energy_demand', data)

    # source scenario
    pop_spec_data = {
        'name': 'population_count',
        'dtype': 'int',
        'dims': ['lad', 'annual'],
        'coords': {
            'lad': sample_dimensions[0]['elements'],
            'annual': annual
        },
        'absolute_range': [0, int(1e12)],
        'expected_range': [0, 100000],
        'unit': 'people'
    }
    store.write_scenario({
        'name': 'population',
        'description': 'Projected annual fertility rates',
        'provides': [
            pop_spec_data
        ],
        'variants': [
            {
                'name': 'low',
                'description': 'Population (Low)',
                'data': {
                    'population_count': 'population_low_count',
                },
            }
        ]
    })
    store.write_scenario_variant_data(
        'population',
        'low',
        DataArray(Spec.from_dict(pop_spec_data), np.array([[1.0], [2.0]])),
        2015
    )

    # source model
    # - test output matches population input to energy demand
    # - test_alt_dims output matches test input to test_convertor
    store.write_model({
        'name': 'test_source',
        'description': '',
        'inputs': [],
        'outputs': [
            {
                'name': 'test',
                'dtype': 'int',
                'dims': ['lad', 'annual'],
                'absolute_range': [0, int(1e12)],
                'expected_range': [0, 100000],
                'unit': 'people'
            },
            {
                'name': 'test_alt_dims',
                'dtype': 'int',
                'dims': ['lad', 'annual'],
                'absolute_range': [0, int(1e12)],
                'expected_range': [0, 100000],
                'unit': 'million people'
            }
        ],
        'parameters': [],
        'interventions': [],
        'initial_conditions': []
    })

    # convertor model
    # - test input matches test output from test_source
    # - test output matches population input to energy demand
    store.write_model({
        'name': 'test_convertor',
        'description': '',
        'inputs': [
            {
                'name': 'test',
                'dtype': 'int',
                'dims': ['lad', 'annual'],
                'absolute_range': [0, int(1e12)],
                'expected_range': [0, 100000],
                'unit': 'million people'
            }
        ],
        'outputs': [
            {
                'name': 'test',
                'dtype': 'int',
                'dims': ['lad', 'annual'],
                'absolute_range': [0, int(1e12)],
                'expected_range': [0, 100000],
                'unit': 'people'
            }
        ],
        'parameters': [],
        'interventions': [],
        'initial_conditions': []
    })

    store.write_sos_model({
        'name': 'energy_demand_only',
        'sector_models': ['energy_demand'],
        'model_dependencies': [],
        'scenario_dependencies': [
            {
                'source': 'population',
                'source_output': 'population_count',
                'sink_input': 'population',
                'sink': 'energy_demand'
            }
        ],
        'narratives': []
    })

    store.write_sos_model({
        'name': 'test_sos_model',
        'sector_models': ['energy_demand', 'test_source'],
        'scenario_dependencies': [],
        'model_dependencies': [
            {
                'source': 'test_source',
                'source_output': 'test',
                'sink_input': 'population',
                'sink': 'energy_demand'
            }
        ],
        'narratives': [
            {
                'name': 'test_narrative',
                'description': 'a narrative config',
                'provides': {
                    'energy_demand': ['smart_meter_savings']
                },
                'variants': [
                    {
                        'name': 'high_tech_dsm',
                        'description': 'High takeup',
                        'data': {'smart_meter_savings': 'filename.csv'}
                    }
                ]
            }
        ]
    })
    parameter_spec = Spec.from_dict(get_sector_model['parameters'][0])
    da = DataArray(parameter_spec, np.array(99))
    store.write_narrative_variant_data(
        'test_sos_model', 'test_narrative', 'high_tech_dsm', da)

    store.write_sos_model({
        'name': 'test_converting_sos_model',
        'sector_models': ['energy_demand', 'test_source', 'test_convertor'],
        'scenario_dependencies': [],
        'model_dependencies': [
            {
                'source': 'test_source',
                'source_output': 'test_alt_dims',
                'sink_input': 'test',
                'sink': 'test_convertor'
            },
            {
                'source': 'test_convertor',
                'source_output': 'test',
                'sink_input': 'population',
                'sink': 'energy_demand'
            }
        ],
        'narratives': []
    })

    store.write_model_run({
        'name': 1,
        'narratives': {},
        'sos_model': 'test_sos_model',
        'scenarios': {}
    })
    store.write_model_run({
        'name': 2,
        'narratives': {},
        'sos_model': 'test_converting_sos_model',
        'scenarios': {}
    })
    store.write_model_run({
        'name': 3,
        'narratives': {},
        'sos_model': 'energy_demand_only',
        'scenarios': {
            'population': 'low'
        }
    })

    return store


class EmptySectorModel(SectorModel):
    """Sector Model implementation
    """

    def simulate(self, data):
        """no-op
        """
        pass


@fixture
def mock_model(mock_store):
    return EmptySectorModel.from_dict(mock_store.read_model('energy_demand'))


@fixture
def mock_model_with_conversion(mock_store):
    return EmptySectorModel.from_dict(mock_store.read_model('test_convertor'))


@fixture
def mock_sos_model(mock_store, mock_model):
    test_source_model = EmptySectorModel.from_dict(mock_store.read_model('test_source'))
    return SosModel.from_dict(mock_store.read_sos_model(
        'test_sos_model'), [mock_model, test_source_model])


class TestDataHandle():
    """Test data handle
    """

    def test_create(self, mock_store, mock_model):
        """should be created with a Store
        """
        DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

    def test_get_data_from_scenario(self, mock_store, mock_model):
        """should allow read access to input data from scenarios
        """
        data_handle = DataHandle(mock_store, 3, 2015, [2015, 2020], mock_model)
        data = np.array([[1.0], [2.0]])

        input_spec = mock_model.inputs['population']
        input_da = DataArray(input_spec, data)

        actual = data_handle.get_data("population")
        assert actual.name == 'population'
        np.testing.assert_equal(actual, input_da)

    def test_get_data_from_model_output(self, mock_store, mock_model):
        """should allow read access to input data from model results
        """
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        data = np.array([[1.0], [2.0]])

        input_spec = mock_model.inputs['population']
        input_da = DataArray(input_spec, data)
        output_spec = copy(input_spec)
        output_spec.name = 'test'  # use source spec name
        output_da = DataArray(output_spec, data)

        mock_store.write_results(
            output_da,
            1,
            'test_source',  # write source model results
            2015,
            None
        )
        actual = data_handle.get_data("population")
        assert output_da.name == 'test'
        assert actual.name == 'population'
        np.testing.assert_equal(actual, input_da)

    def test_get_data_with_conversion(self, mock_store, mock_model_with_conversion):
        """should convert liters to milliliters (1 -> 0.001)
        """
        modelrun_name = 2
        data_handle = DataHandle(
            mock_store, modelrun_name, 2015, [2015, 2020], mock_model_with_conversion)
        data = np.array([[0.001], [0.002]])

        spec = copy(mock_model_with_conversion.inputs['test'])
        spec.name = 'test_alt_dims'  # use source spec name
        da = DataArray(spec, data)

        mock_store.write_results(
            da,
            modelrun_name,
            'test_source',  # write results as though from source
            2015,
            None
        )
        actual = data_handle.get_data("test")

        expected = DataArray(spec, data)

        np.testing.assert_equal(actual, expected)

    def test_get_base_timestep_data(self, mock_store, mock_model):
        """should allow read access to input data from base timestep
        """
        data_handle = DataHandle(mock_store, 1, 2025, [2015, 2020, 2025], mock_model)
        data = np.array([[1.0], [2.0]])

        spec = copy(mock_model.inputs['population'])
        spec.name = 'test'  # use source spec name
        da = DataArray(spec, data)

        mock_store.write_results(
            da,
            1,
            'test_source',  # write source model results
            2015,  # base timetep
            None
        )

        actual = data_handle.get_base_timestep_data("population")
        expected = DataArray(spec, data)
        np.testing.assert_equal(actual, expected)

    def test_get_base_timestep_data_before_model_run(self, mock_store, mock_model):
        """Prior to a model run, there is no current timestep

        This should allow read access to input data from base timestep
        """
        data_handle = DataHandle(mock_store, 1, None, [2015, 2020, 2025], mock_model)
        data = np.array([[1.0], [2.0]])

        spec = copy(mock_model.inputs['population'])
        spec.name = 'test'  # use source spec name
        da = DataArray(spec, data)

        mock_store.write_results(
            da,
            1,
            'test_source',  # write source model results
            2015,  # base timetep
            None
        )

        actual = data_handle.get_base_timestep_data("population")
        expected = DataArray(spec, data)
        np.testing.assert_equal(actual, expected)

    def test_get_previous_timestep_data(self, mock_store, mock_model):
        """should allow read access to input data from previous timestep
        """
        data_handle = DataHandle(mock_store, 1, 2025, [2015, 2020, 2025], mock_model)
        data = np.random.rand(*mock_model.inputs['population'].shape)
        spec = copy(mock_model.inputs['population'])
        spec.name = 'test'  # use source spec name
        da = DataArray(spec, data)

        mock_store.write_results(
            da,
            1,
            'test_source',  # write source model results
            2020,  # previous timetep
            None
        )
        actual = data_handle.get_previous_timestep_data("population")

        expected = DataArray(spec, data)

        assert actual == expected

    def test_get_data_with_square_brackets(self, mock_store, mock_model):
        """should allow dict-like read access to input data
        """
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        data = np.random.rand(*mock_model.inputs['population'].shape)

        spec = copy(mock_model.inputs['population'])
        spec.name = 'test'  # use source spec name
        da = DataArray(spec, data)

        mock_store.write_results(
            da,
            1,
            'test_source',  # write source model results
            2015,  # current timetep
            None
        )
        actual = data_handle["population"]
        assert actual == da

    def test_set_data(self, mock_store, mock_model_with_conversion):
        """should allow write access to output data
        """
        data = np.array([[1.0], [4.0]])
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model_with_conversion)

        spec = mock_model_with_conversion.outputs['test']
        da = DataArray(spec, data)

        data_handle.set_results("test", data)
        actual = mock_store.read_results(
            1,
            'test_convertor',  # read results from model
            mock_model_with_conversion.outputs['test'],
            2015,
            None
        )
        np.testing.assert_equal(actual.as_ndarray(), data)
        assert actual == da

    def test_set_data_wrong_shape(self, mock_store, mock_model_with_conversion):
        """should allow write access to output data
        """
        data = np.array([[1.0, 1.0]])  # regions is 1, intervals is 1 not 2

        spec = mock_model_with_conversion.outputs['test']
        with raises(SmifDataMismatchError) as ex:
            DataArray(spec, data)

        msg = "Data shape (1, 2) does not match spec " \
              "(2, 1)"
        assert msg in str(ex)

    def test_set_data_with_square_brackets(self, mock_store, mock_model):
        """should allow dict-like write access to output data
        """
        data = np.random.rand(2, 8)
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        spec = mock_model.outputs['gas_demand']
        da = DataArray(spec, data)

        data_handle["gas_demand"] = data
        actual = mock_store.read_results(
            1,
            'energy_demand',  # read results from model
            mock_model.outputs['gas_demand'],
            2015
        )
        np.testing.assert_equal(actual.as_ndarray(), data)
        assert actual == da

    def test_set_data_with_square_brackets_raises(self, mock_store, mock_model):
        """should allow dict-like write access to output data
        """
        data = np.random.rand(2, 8)
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        spec = mock_model.outputs['gas_demand']
        da = DataArray(spec, data)

        with raises(TypeError) as err:
            data_handle["gas_demand"] = da

        assert "Pass in a numpy array" in str(err)


class TestDataHandleState():
    """Test handling of initial conditions, decision interventions and intervention state.
    """

    def test_get_state(self, mock_store, mock_model):
        """should get decision module state for given timestep/decision_iteration

        A call to ``get_state`` method on the data handle calls the read_state
        method of the store with arguments for model run name, current
        timestep and decision iteration.
        """
        mock_store.read_state = Mock(return_value=[{'name': 'test', 'build_year': 2010}])
        mock_store.write_interventions('energy_demand', [{
            'name': 'test',
            'capital_cost': {'value': 2500, 'unit': '£/GW'}
        }])
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        expected = [{
            'name': 'test',
            'build_year': 2010
        }]
        actual = data_handle.get_state()
        mock_store.read_state.assert_called_with(1, 2015, None)
        assert actual == expected

    def test_get_interventions_for_sector_model(self, mock_store, mock_model):
        """

        A call to the ``get_current_interventions`` method of the data_handle
        returns a dict of interventions from the current state filtered by the
        current model

        """
        state = [
            {'name': 'water_asset_a', 'build_year': 2010},
            {'name': 'water_asset_b', 'build_year': 2015}
        ]
        mock_store.write_state(state, 1, 2015, None)
        # mock_store._strategies[1] = []

        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        actual = data_handle.get_current_interventions()
        expected = {
            'water_asset_a':
                {'build_year': 2010,
                 'capacity': 50,
                 'location': None,
                 'sector': ''},
            'water_asset_b':
                {'build_year': 2015,
                 'capacity': 150,
                 'location': None,
                 'sector': ''}
        }
        assert actual == expected

    def test_interventions_sector_model_ignore_unrecog(self, mock_store, mock_model):
        """Ignore unrecognised interventions

        Interventions that are listed in state, but not included in the
        intervention list are ignored

        """
        state = [
            {'name': 'water_asset_a', 'build_year': 2010},
            {'name': 'energy_asset_unexpected', 'build_year': 2015}
        ]

        mock_store.write_state(state, 1, 2015, None)

        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        actual = data_handle.get_current_interventions()
        expected = {'water_asset_a': {
            'build_year': 2010,
            'capacity': 50,
            'location': None,
            'sector': ''
        }
        }
        assert actual == expected

    def test_pass_none_to_timestep_raises(self, mock_store, mock_model):
        data_handle = DataHandle(mock_store, 1, None, [2015, 2020], mock_model)
        with raises(ValueError):
            data_handle.get_state()


class TestDataHandleTimesteps():
    """Test timestep helper properties
    """

    def test_current_timestep(self, mock_model, mock_store):
        """should return current timestep
        """
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        assert data_handle.current_timestep == 2015

    def test_base_timestep(self, mock_model, mock_store):
        """should return first timestep in list
        """
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        assert data_handle.base_timestep == 2015

    def test_previous_timestep(self, mock_model, mock_store):
        """should return previous timestep from list
        """
        data_handle = DataHandle(mock_store, 1, 2020, [2015, 2020], mock_model)
        assert data_handle.previous_timestep == 2015

    def test_previous_timestep_error(self, mock_model, mock_store):
        """should raise error if there's no previous timestep in the list
        """
        data_handle = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        with raises(SmifTimestepResolutionError) as ex:
            data_handle.previous_timestep
        assert 'no previous timestep' in str(ex)


class TestDataHandleGetResults:

    def test_get_results_sectormodel(self, mock_store,
                                     mock_model):
        """Get results from a sector model
        """
        store = mock_store
        spec = mock_model.outputs['gas_demand']

        data = np.random.rand(2, 8)
        da = DataArray(spec, data)

        store.write_results(da, 1, 'energy_demand', 2010)

        dh = DataHandle(mock_store, 1, 2010, [2010], mock_model)
        actual = dh.get_results('gas_demand')
        assert actual == DataArray(spec, data)

    def test_get_results_no_output_sector(self, mock_store,
                                          mock_model):

        with raises(KeyError):
            dh = DataHandle(mock_store, 1, 2010, [2010], mock_model)
            dh.get_results('no_such_output')


class TestDataHandleGetParameters:

    def test_load_parameter_defaults(self, mock_store, mock_model):

        dh = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        actual = dh.get_parameter('smart_meter_savings')
        spec = Spec.from_dict(
            {
                'name': 'smart_meter_savings',
                'description': "Difference in floor area per person"
                               "in end year compared to base year",
                'absolute_range': [0, float('inf')],
                'expected_range': [0.5, 2],
                'unit': '%',
                'dtype': 'float'
            })
        expected = DataArray(spec, np.array(42, dtype=float))

        assert actual == expected

    def test_load_parameters_override(self, mock_store, mock_model):

        mock_store.update_model_run(1, {
            'name': 1,
            'narratives': {'test_narrative': ['high_tech_dsm']},
            'sos_model': 'test_sos_model',
            'scenarios': {}})
        dh = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        actual = dh.get_parameter('smart_meter_savings')
        spec = Spec.from_dict(
            {
                'name': 'smart_meter_savings',
                'description': "Difference in floor area per person"
                               "in end year compared to base year",
                'absolute_range': [0, float('inf')],
                'expected_range': [0.5, 2],
                'unit': '%',
                'dtype': 'float'
            })
        expected = DataArray(spec, np.array(99))

        assert actual == expected

    def test_load_parameters_override_ordered(self, mock_store, mock_model):
        """Parameters in a narrative variants listed later override parameters
        contained in earlier variants
        """
        sos_model = mock_store.read_sos_model('test_sos_model')
        sos_model['narratives'] = [{
            'name': 'test_narrative',
            'description': 'a narrative config',
            'sos_model': 'test_sos_model',
            'provides': {'energy_demand': ['smart_meter_savings']},
            'variants': [
                {
                    'name': 'first_variant',
                    'description': 'This variant should be overridden',
                    'data': {'smart_meter_savings': 'filename.csv'}},
                {
                    'name': 'second_variant',
                    'description': 'This variant should override the first',
                    'data': {'smart_meter_savings': 'filename.csv'}}
            ]
        }]
        mock_store.update_sos_model('test_sos_model', sos_model)

        mock_store.update_model_run(1, {
            'name': 1,
            'narratives': {'test_narrative': ['first_variant',
                                              'second_variant']},
            'sos_model': 'test_sos_model',
            'scenarios': {}})

        parameter_spec = Spec(
            name='smart_meter_savings',
            dtype='float',
            unit='%'
        )
        first_variant = DataArray(parameter_spec, np.array(1))
        mock_store.write_narrative_variant_data(
            'test_sos_model', 'test_narrative', 'first_variant', first_variant)

        second_variant = DataArray(parameter_spec, np.array(2))
        mock_store.write_narrative_variant_data(
            'test_sos_model', 'test_narrative', 'second_variant', second_variant)

        dh = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)

        actual = dh.get_parameter('smart_meter_savings')

        assert actual == second_variant

    def test_load_parameters_partial_override(self, mock_store, mock_model):
        # add parameter to model
        param_spec = Spec(
            name='multi_savings',
            description='The savings from various technologies',
            abs_range=(0, 100),
            exp_range=(3, 10),
            dims=['technology_type'],
            coords={'technology_type': ['water_meter', 'electricity_meter']},
            unit='%',
            dtype='float'
        )
        mock_model.add_parameter(param_spec)
        mock_store.update_model(mock_model.name, mock_model.as_dict())

        # default values
        param_defaults = DataArray(param_spec, np.array([3, 3]))
        mock_store.write_model_parameter_default(
            mock_model.name, param_spec.name, param_defaults)

        # add narrative to sosmodel
        sos_model = mock_store.read_sos_model('test_sos_model')
        sos_model['narratives'] = [{
            'name': 'test_narrative',
            'description': 'a narrative config',
            'sos_model': 'test_sos_model',
            'provides': {'energy_demand': ['multi_savings']},
            'variants': [
                {
                    'name': 'low_multi_save',
                    'description': 'Low values',
                    'data': {'multi_savings': 'multi_savings.csv'}
                }
            ]
        }]
        mock_store.update_sos_model('test_sos_model', sos_model)

        # narrative values - one NaN i.e. not overridden
        param_narrative = DataArray(param_spec, np.array([np.nan, 99]))
        mock_store.write_narrative_variant_data(
            'test_sos_model', 'test_narrative', 'low_multi_save', param_narrative)

        # expect combination
        expected = DataArray(param_spec, np.array([3, 99]))

        mock_store.update_model_run(1, {
            'name': 1,
            'narratives': {'test_narrative': ['low_multi_save']},
            'sos_model': 'test_sos_model',
            'scenarios': {}
        })

        dh = DataHandle(mock_store, 1, 2015, [2015, 2020], mock_model)
        actual = dh.get_parameter('multi_savings')

        assert actual == expected


class TestDataHandleCoefficients:
    """Tests the interface for reading and writing coefficients
    """
    def test_read_coefficient_raises(self, mock_store, mock_model):
        """Reading coefficients that aren't cached should raise SmifDataNotFoundError
        """
        store = mock_store
        dh = DataHandle(store, 1, 2010, [2010], mock_model)

        with raises(SmifDataNotFoundError):
            dh.read_coefficients('from_dim', 'to_dim')

    def test_write_coefficients(self, mock_store, mock_model):
        """Writing coefficients should succeed
        """
        store = mock_store
        data = np.zeros(4)

        dh = DataHandle(store, 1, 2010, [2010], mock_model)
        dh.write_coefficients('from_dim', 'to_dim', data)

    def test_read_coefficient_(self, mock_store, mock_model):
        """Reading written coefficients should return the cached data
        """
        store = mock_store
        dh = DataHandle(store, 1, 2010, [2010], mock_model)
        data = np.zeros(4)
        dh.write_coefficients('from_dim', 'to_dim', data)

        actual = dh.read_coefficients('from_dim', 'to_dim')
        np.testing.assert_equal(actual, np.array([0, 0, 0, 0]))

    def test_read_unit_definitions(self, mock_store, mock_model):
        """Reading list of unit definition strings
        """
        store = mock_store
        dh = DataHandle(store, 1, 2010, [2010], mock_model)
        actual = dh.read_unit_definitions()
        assert actual == [
            'people',
            'million people = 1000000 * people'
        ]


class TestResultsHandle:
    """Get results from any model
    """

    def test_get_results_sos_model(self, mock_store, mock_model, mock_sos_model):
        """Get results from a sector model within a sos model
        """
        store = mock_store
        spec = mock_model.outputs['gas_demand']

        da = DataArray(spec, np.random.rand(2, 8))

        store.write_results(da, 'test_modelrun', 'energy_demand', 2010, None)

        dh = ResultsHandle(mock_store, 'test_modelrun', mock_sos_model, 2010)
        actual = dh.get_results('energy_demand', 'gas_demand', 2010, None)
        spec = mock_model.outputs['gas_demand']
        assert actual == da

    def test_get_results_no_output_sos(self, mock_store, mock_sos_model):
        with raises(KeyError):
            dh = ResultsHandle(mock_store, 'test_modelrun', mock_sos_model, 2010)
            dh.get_results('energy_demand', 'no_such_output', 2010, None)

    def test_get_results_wrong_name_sos(self, mock_store, mock_sos_model):
        with raises(KeyError):
            dh = ResultsHandle(mock_store, 'test_modelrun', mock_sos_model, 2010)
            dh.get_results('no_such_model', 'test_output', 2010, None)

    def test_get_results_not_exists(self, mock_store, mock_sos_model, mock_model):
        store = mock_store

        spec = mock_model.outputs['gas_demand']
        da = DataArray(spec, np.random.rand(2, 8))

        store.write_results(da, 'test_modelrun', 'energy_demand', 2010, None)
        dh = ResultsHandle(store, 'test_modelrun', mock_sos_model, 2100)
        with raises(SmifDataError):
            dh.get_results('energy_demand', 'gas_demand', 2099, None)



"""Test unit adaptor
"""
from unittest.mock import Mock

import numpy as np
from smif.convert.unit import UnitAdaptor
from smif.data_layer.data_array import DataArray
from smif.metadata import Spec


def test_convert_unit():
    """Convert SI units
    """
    data_handle = Mock()
    data = np.array([1], dtype=float)

    from_spec = Spec(
        name='test_variable',
        dtype='float',
        unit='liter'
    )

    data_array = DataArray(from_spec, data)

    data_handle.get_data = Mock(return_value=data_array)

    adaptor = UnitAdaptor('test-ml-l')
    adaptor.add_input(from_spec)
    adaptor.add_output(Spec(
        name='test_variable',
        dtype='float',
        unit='milliliter'
    ))
    adaptor.simulate(data_handle)

    actual = data_handle.set_results.call_args[0][1]
    expected = np.array([1000], dtype=float)
    np.testing.assert_allclose(actual, expected)


def test_convert_custom():
    """Convert custom units
    """
    data_handle = Mock()
    data = np.array([0.18346346], dtype=float)

    from_spec = Spec(
        name='test_variable',
        dtype='float',
        unit='mcm'
    )

    to_spec = Spec(
        name='test_variable',
        dtype='float',
        unit='GW'
    )

    data_array = DataArray(from_spec, data)

    data_handle.get_data = Mock(return_value=data_array)
    data_handle.read_unit_definitions = Mock(return_value=['mcm = 10.901353 * GW'])

    adaptor = UnitAdaptor('test-mcm-GW')
    adaptor.add_input(from_spec)
    adaptor.add_output(to_spec)
    adaptor.before_model_run(data_handle)  # must have run before_model_run to register units
    adaptor.simulate(data_handle)

    actual = data_handle.set_results.call_args[0][1]
    expected = np.array([2], dtype=float)
    np.testing.assert_allclose(actual, expected)



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
from copy import copy
from unittest import TestCase
from unittest.mock import Mock, PropertyMock

import numpy as np
from pytest import fixture, raises
from smif.exception import SmifDataMismatchError, SmifValidationError
from smif.metadata import RelativeTimestep, Spec
from smif.model.dependency import Dependency
from smif.model.model import ScenarioModel
from smif.model.sector_model import SectorModel
from smif.model.sos_model import SosModel


class EmptySectorModel(SectorModel):
    """Simulate nothing
    """
    def simulate(self, data):
        return data


@fixture(scope='function')
def empty_sector_model():
    """SectorModel ready for customisation
    """
    return EmptySectorModel('test')


@fixture(scope='function')
def scenario_model():
    """ScenarioModel providing precipitation
    """
    model = ScenarioModel('climate')
    model.add_output(
        Spec.from_dict({
            'name': 'precipitation',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'ml'
        })
    )
    model.add_output(
        Spec.from_dict({
            'name': 'reservoir_level',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'ml'
        })
    )
    model.scenario = 'UKCP09 High emissions'
    return model


@fixture(scope='function')
def sector_model():
    """SectorModel requiring precipitation and cost, providing water
    """
    model = EmptySectorModel('water_supply')
    model.add_input(
        Spec.from_dict({
            'name': 'precipitation',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'ml'
        })
    )
    model.add_input(
        Spec.from_dict({
            'name': 'reservoir_level',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'ml'
        })
    )
    model.add_input(
        Spec.from_dict({
            'name': 'rGVA',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'million GBP'
        })
    )
    model.add_output(
        Spec.from_dict({
            'name': 'water',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'Ml'
        })
    )
    model.add_output(
        Spec.from_dict({
            'name': 'reservoir_level',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'ml'
        })
    )
    model.add_parameter(
        Spec.from_dict({
            'name': 'test_parameter',
            'description': 'a dummy parameter to test narratives',
            'dims': ['national'],
            'coords': {'national': ['GB']},
            'abs_range': (0.5, 2),
            'exp_range': (0.5, 2),
            'dtype': 'float',
            'default': [[1.]],
            'unit': '%'
        })
    )
    return model


@fixture(scope='function')
def economic_model():
    """SectorModel requiring precipitation and cost, providing water
    """
    model = EmptySectorModel('economic_model')
    model.add_output(
        Spec.from_dict({
            'name': 'gva',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'million GBP'
        })
    )
    return model


@fixture
def energy_model():
    class EnergyModel(SectorModel):
        """
        electricity_demand_input -> fluffiness
        """

        def simulate(self, data_handle):
            """Mimics the running of a sector model
            """
            fluff = data_handle['electricity_demand_input']
            data_handle['fluffiness'] = fluff * 0.819
            return data_handle

    energy_model = EnergyModel('energy_model')
    energy_model.add_input(
        Spec(
            name='electricity_demand_input',
            dims=['LSOA'],
            coords={'LSOA': ['E090001', 'E090002']},
            dtype='float',
        )
    )
    energy_model.add_output(
        Spec(
            name='fluffiness',
            dims=['LSOA'],
            coords={'LSOA': ['E090001', 'E090002']},
            dtype='float',
        )
    )

    return energy_model


@fixture(scope='function')
def sos_model(sector_model, scenario_model, economic_model):
    """SosModel with one scenario and one sector model
    """
    model = SosModel('test_sos_model')
    model.add_model(scenario_model)
    model.add_model(economic_model)
    model.add_model(sector_model)
    model.add_dependency(scenario_model, 'precipitation', sector_model, 'precipitation')
    model.add_dependency(economic_model, 'gva', sector_model, 'rGVA')
    return model


@fixture(scope='function')
def scenario_only_sos_model_dict():
    """Config for a SosModel with one scenario
    """
    return {
        'name': 'test_sos_model',
        'description': 'Readable description of the sos model',
        'scenarios': [
            {
                'name': 'climate',
                'scenario': 'UKCP09 High emissions',
                'outputs': [
                    {
                        'name': 'precipitation',
                        'dims': ['LSOA'],
                        'coords': {'LSOA': [1, 2, 3]},
                        'dtype': 'float',
                        'unit': 'ml'
                    },
                    {
                        'name': 'reservoir_level',
                        'dims': ['LSOA'],
                        'coords': {'LSOA': [1, 2, 3]},
                        'dtype': 'float',
                        'unit': 'ml'
                    }
                ]
            }
        ],
        'sector_models': [],
        'model_dependencies': [],
        'scenario_dependencies': []
    }


@fixture(scope='function')
def sos_model_dict(scenario_only_sos_model_dict):
    """Config for a SosModel with one scenario and one sector model
    """
    config = scenario_only_sos_model_dict
    config['sector_models'] = [
        {
            'name': 'economic_model',
            'inputs': [],
            'parameters': [],
            'outputs': [
                {
                    'name': 'gva',
                    'dims': ['LSOA'],
                    'coords': {'LSOA': [1, 2, 3]},
                    'dtype': 'float',
                    'unit': 'million GBP'
                }
            ]
        },
        {
            'name': 'water_supply',
            'inputs': [
                {
                    'name': 'precipitation',
                    'dims': ['LSOA'],
                    'coords': {'LSOA': [1, 2, 3]},
                    'dtype': 'float',
                    'unit': 'ml'
                },
                {
                    'name': 'rGVA',
                    'dims': ['LSOA'],
                    'coords': {'LSOA': [1, 2, 3]},
                    'dtype': 'float',
                    'unit': 'million GBP'
                },
                {
                    'name': 'reservoir_level',
                    'dims': ['LSOA'],
                    'coords': {'LSOA': [1, 2, 3]},
                    'dtype': 'float',
                    'unit': 'ml'
                }
            ],
            'parameters': [],
            'outputs': [
                {
                    'name': 'water',
                    'dims': ['LSOA'],
                    'coords': {'LSOA': [1, 2, 3]},
                    'dtype': 'float',
                    'unit': 'Ml'
                },
                {
                    'name': 'reservoir_level',
                    'dims': ['LSOA'],
                    'coords': {'LSOA': [1, 2, 3]},
                    'dtype': 'float',
                    'unit': 'ml'
                }
            ]
        }
    ]
    config['scenario_dependencies'] = [
        {
            'source': 'climate',
            'source_output': 'precipitation',
            'sink_input': 'precipitation',
            'sink': 'water_supply',
            'timestep': 'CURRENT'
        },
        {
            'source': 'climate',
            'source_output': 'reservoir_level',
            'sink_input': 'reservoir_level',
            'sink': 'water_supply',
            'timestep': 'CURRENT'
        }
    ]
    config['model_dependencies'] = [
        {
            'source': 'economic_model',
            'source_output': 'gva',
            'sink_input': 'rGVA',
            'sink': 'water_supply',
            'timestep': 'CURRENT'
        },
        {
            'source': 'water_supply',
            'source_output': 'reservoir_level',
            'sink_input': 'reservoir_level',
            'sink': 'water_supply',
            'timestep': 'PREVIOUS'
        }
    ]
    return config


class TestSosModel():
    """Construct from config or compose from objects
    """
    def test_construct(self, sos_model_dict, scenario_model, sector_model, economic_model):
        """Constructing from config of the form::

            {
                'name': 'sos_model_name',
                'description': 'friendly description of the sos model',
                'dependencies': [
                    {
                        'source': str (Model.name),
                        'source_output': str (Metadata.name),
                        'sink': str (Model.name),
                        'sink_output': str (Metadata.name)
                    }
                ]
            }

        With list of child SectorModel/ScenarioModel instances passed in alongside.
        """
        sos_model = SosModel.from_dict(
            sos_model_dict, [scenario_model, sector_model, economic_model])

        assert isinstance(sos_model, SosModel)
        assert list(sos_model.scenario_models) == [scenario_model]
        assert isinstance(sos_model.get_model('climate'), ScenarioModel)
        assert sos_model.sector_models == [sector_model, economic_model] or \
            sos_model.sector_models == [economic_model, sector_model]
        assert isinstance(sos_model.get_model('economic_model'), SectorModel)
        assert isinstance(sos_model.get_model('water_supply'), SectorModel)

    def test_optional_description(self, sos_model_dict):
        """Default to empty description
        """
        del sos_model_dict['description']
        sos_model = SosModel.from_dict(sos_model_dict)
        assert sos_model.description == ''

    def test_dependencies_fields(self, sos_model_dict, scenario_model, sector_model,
                                 economic_model):
        """Compose dependencies from scenario- and model- fields
        """
        sos_model_dict['scenario_dependencies'] = [
            {
                'source': 'climate',
                'source_output': 'precipitation',
                'sink_input': 'precipitation',
                'sink': 'water_supply',
                'timestep': 'CURRENT'
            },
            {
                'source': 'climate',
                'source_output': 'reservoir_level',
                'sink_input': 'reservoir_level',
                'sink': 'water_supply',
                'timestep': 'CURRENT'
            }
        ]
        sos_model_dict['model_dependencies'] = [
            {
                'source': 'economic_model',
                'source_output': 'gva',
                'sink_input': 'rGVA',
                'sink': 'water_supply',
                'timestep': 'CURRENT'
            },
            {
                'source': 'water_supply',
                'source_output': 'reservoir_level',
                'sink_input': 'reservoir_level',
                'sink': 'water_supply',
                'timestep': 'PREVIOUS'
            }
        ]
        sos_model = SosModel.from_dict(
            sos_model_dict, [scenario_model, sector_model, economic_model])
        actual = sos_model.as_dict()
        TestCase().assertCountEqual(
            actual['scenario_dependencies'], sos_model_dict['scenario_dependencies'])
        TestCase().assertCountEqual(
            actual['model_dependencies'], sos_model_dict['model_dependencies'])

    def test_compose(self, sos_model):
        with raises(AssertionError) as ex:
            sos_model.add_model(SosModel('test'))
        msg = "Only Models can be added to a SosModel (and SosModels cannot be nested)"
        assert msg in str(ex)

    def test_as_dict(self, sos_model):
        """as_dict correctly returns configuration as a dictionary, with child models as_dict
        similarly
        """
        actual = sos_model.as_dict()
        del actual['scenario_dependencies']
        del actual['model_dependencies']
        expected = {
            'name': 'test_sos_model',
            'description': '',
            'scenarios': ['climate'],
            'sector_models': ['economic_model', 'water_supply'],
            'narratives': {}
        }
        assert actual == expected

    def test_add_dependency(self, empty_sector_model):
        """Add models, connect via dependency
        """
        spec = Spec(
            name='hourly_value',
            dims=['hours'],
            coords={'hours': range(24)},
            dtype='int'
        )
        sink_model = copy(empty_sector_model)
        sink_model.add_input(spec)

        source_model = copy(empty_sector_model)
        source_model.add_output(spec)

        sos_model = SosModel('test')
        sos_model.add_model(source_model)
        sos_model.add_model(sink_model)

        sos_model.add_dependency(source_model, 'hourly_value', sink_model, 'hourly_value')

    def test_before_model_run_fails(self, sos_model):
        """Before model run should raise
        """
        data_handle = Mock()
        with raises(AttributeError):
            sos_model.before_model_run(data_handle)

    def test_simulate_fails(self, sos_model):
        """Simulate should raise
        """
        data_handle = Mock()
        with raises(AttributeError):
            sos_model.simulate(data_handle)


class TestSosModelProperties():
    """SosModel has inputs, outputs, parameters

    Convergence settings (should perhaps move to runner)
    """

    def test_model_inputs(self, sos_model):
        spec = sos_model.get_model('water_supply').inputs['precipitation']
        assert isinstance(spec, Spec)

    def test_model_outputs(self, sos_model):
        spec = sos_model.get_model('water_supply').outputs['water']
        assert isinstance(spec, Spec)


class TestSosModelDependencies(object):
    """SosModel can represent data flow as defined by model dependencies
    """
    def test_simple_dependency(self, sos_model):
        """Dependencies
        """
        scenario = sos_model.get_model('climate')
        model = sos_model.get_model('water_supply')
        economic_model = sos_model.get_model('economic_model')

        actual = list(sos_model.dependencies)
        expected = [
            Dependency(
                scenario,
                scenario.outputs['precipitation'],
                model,
                model.inputs['precipitation']
            ),
            Dependency(
                economic_model,
                economic_model.outputs['gva'],
                model,
                model.inputs['rGVA']
            )
        ]
        assert actual == expected or actual == list(reversed(expected))

        actual = list(sos_model.scenario_dependencies)
        expected = [
            Dependency(
                scenario,
                scenario.outputs['precipitation'],
                model,
                model.inputs['precipitation']
            )
        ]
        assert actual == expected

        actual = list(sos_model.model_dependencies)
        expected = [
            Dependency(
                economic_model,
                economic_model.outputs['gva'],
                model,
                model.inputs['rGVA']
            )
        ]
        assert actual == expected

    def test_dependency_timestep(self, sos_model, scenario_model, sector_model):
        # add self-dependency on previous timestep output
        sos_model.add_dependency(
            sector_model, 'reservoir_level', sector_model, 'reservoir_level',
            timestep=RelativeTimestep.PREVIOUS)
        # add dependency on scenario (to satisfy initial timestep input requirement)
        sos_model.add_dependency(
            scenario_model, 'reservoir_level',
            sector_model, 'reservoir_level'
        )

    def test_dependency_duplicate(self, sos_model, scenario_model, sector_model):
        with raises(SmifValidationError) as ex:
            sos_model.add_dependency(
                scenario_model, 'precipitation',
                sector_model, 'precipitation')
        assert "Could not add dependency: input 'precipitation' already provided" in str(ex)

    def test_dependency_not_present(self, sos_model, scenario_model, energy_model):
        """Should fail with missing input/output
        """
        sos_model.add_model(energy_model)
        with raises(SmifValidationError) as ex:
            sos_model.add_dependency(
                scenario_model, 'not_present', energy_model, 'electricity_demand_input')
        msg = "Output 'not_present' is not defined in '{}'".format(scenario_model.name)
        assert msg in str(ex)

        with raises(SmifValidationError) as ex:
            sos_model.add_dependency(
                scenario_model, 'precipitation', energy_model, 'incorrect_name')
        msg = "Input 'incorrect_name' is not defined in '{}'".format(energy_model.name)
        assert msg in str(ex)

    def test_dependency_model_not_exist(self, sos_model, scenario_model, energy_model):
        """Should fail with a SmifConfigurationError
        """

        missing_sink = Mock()
        type(missing_sink).name = PropertyMock(return_value='test_sink_model')

        with raises(SmifValidationError) as ex:
            sos_model.add_dependency(
                scenario_model, 'precipitation', missing_sink, 'does not matter'
            )
        msg = "Sink model 'test_sink_model' does not exist in list of models"
        assert msg in str(ex)

        missing_source = Mock()
        type(missing_source).name = PropertyMock(return_value='test_source_model')

        with raises(SmifValidationError) as ex:
            sos_model.add_dependency(
                missing_source, 'does not matter', energy_model, 'electricity_demand_input'
            )
        msg = "Source model 'test_source_model' does not exist in list of models"
        assert msg in str(ex)

    def test_data_not_present(self, sos_model_dict, sector_model):
        """Raise a NotImplementedError if an input is defined but no dependency links
        it to a data source
        """
        sos_model_dict['scenario_dependencies'] = []
        sos_model_dict['model_dependencies'] = []
        with raises(NotImplementedError):
            SosModel.from_dict(sos_model_dict, [sector_model])

    def test_undefined_unit_conversion(self, sos_model_dict, sector_model, scenario_model,
                                       economic_model):
        """Error on invalid dependency
        """
        sector_model.inputs['precipitation'] = Spec(
            name='precipitation',
            dims=['LSOA'],
            coords={'LSOA': [1, 2, 3]},
            dtype='float',
            unit='incompatible'
        )

        with raises(ValueError) as ex:
            SosModel.from_dict(sos_model_dict, [sector_model, scenario_model, economic_model])
        assert "ml!=incompatible" in str(ex)

    def test_invalid_unit_conversion(self, sos_model_dict, sector_model, scenario_model,
                                     economic_model):
        """Error on invalid dependency
        """
        scenario_model.outputs['precipitation'] = Spec(
            name='precipitation',
            dims=['LSOA'],
            coords={'LSOA': [1, 2, 3]},
            dtype='float',
            unit='meter'
        )

        with raises(ValueError) as ex:
            SosModel.from_dict(sos_model_dict,
                               [sector_model, scenario_model, economic_model])
        assert "meter!=ml" in str(ex.value)

    def test_scenario_variant_not_present(self, sos_model_dict, scenario_model,
                                          economic_model):
        """Error on sector model not provided
        """
        with raises(SmifDataMismatchError) as ex:
            SosModel.from_dict(sos_model_dict, [scenario_model, economic_model])
        assert "SectorModel or ScenarioModel sink `water_supply` required " + \
               "by dependency `economic_model (gva) - water_supply (rGVA)` " + \
               "was not provided by the builder" in str(ex.value)

    def test_scenario_model_not_provided(self, sos_model_dict, sector_model, economic_model):
        """Error on scenario not provided
        """
        with raises(SmifDataMismatchError) as ex:
            SosModel.from_dict(sos_model_dict, [sector_model, economic_model])
        assert "SectorModel or ScenarioModel source `climate` required by " + \
               "dependency `climate (precipitation) - water_supply (precipitation)` " + \
               "was not provided by the builder" in str(ex.value)


class TestNarratives:

    @fixture(scope='function')
    def narrative(self):
        narrative = {
            'name': 'test_narrative',
            'description': 'a narrative config',
            'provides': {'water_supply': ['test_parameter']},
            'variants': [{
                'name': 'high_tech_dsm',
                'description': 'High takeup',
                'data': {'test_parameter': np.array([[99]])}}]
                    }
        return narrative

    def test_add_narrative(self, sos_model, narrative):

        sos_model.add_narrative(narrative)
        actual = sos_model.narratives['test_narrative']
        expected = narrative
        assert actual == expected

    def test_add_narrative_raises_for_incorrect_parameter(self, sos_model, narrative):

        narrative['provides'] = {'water_supply': ['no_such_parameter']}

        with raises(SmifDataMismatchError) as err:
            sos_model.add_narrative(narrative)

        assert "Parameter 'no_such_parameter' does not exist in 'water_supply'" in str(err)

    def test_add_narrative_raises_for_wrong_model(self, sos_model, narrative):

        narrative['provides'] = {'not_a_model': ['test_parameter']}

        with raises(SmifDataMismatchError) as err:
            sos_model.add_narrative(narrative)

        assert "'not_a_model' does not exist in 'test_sos_model'" in str(err)

    def test_narratives_in_as_dict(self, sos_model, narrative):

        sos_model.add_narrative(narrative)
        actual = sos_model.as_dict()
        assert 'narratives' in actual

        actual_narratives = actual['narratives']
        expected_narratives = narrative
        assert actual_narratives == {'test_narrative': expected_narratives}

    def test_get_parameter_specs_of_narrative(self, sos_model):

        pass



"""Holds custom smif exception hierarchy

Exception
  +-- SmifException
        +-- SmifDataError
              +-- SmifDataNotFoundError
              +-- SmifDataExistsError
              +-- SmifDataMismatchError
              +-- SmifDataReadError
              +-- SmifDataInputError
        +-- SmifModelRunError
        +-- SmifValidationError
"""


class SmifException(Exception):
    """The base class for all errors raised in smif
    """
    pass


class SmifDataError(SmifException):
    """Errors raised by the Store
    """
    pass


class SmifDataNotFoundError(SmifDataError):
    """Raise when some data is not found
    """
    pass


class SmifDataExistsError(SmifDataError):
    """Raise when some data is found unexpectedly
    """
    pass


class SmifDataMismatchError(SmifDataError):
    """Raise when some data doesn't match the context

    E.g. when updating an object by id, the updated object's id must match
    the id provided separately.
    """
    pass


class SmifDataReadError(SmifDataError):
    """Raise when unable to read data

    E.g. unable to handle file type or connect to database
    """
    pass


class SmifDataInputError(SmifDataError):
    """Raise when unable to write data because it does not meet specification
    and can be addressed to a specific (user-interface) input field

    E.g.
    - component: description
    - unable to write a description shorter than 5 characters
    - We require a description so you can identify your system-of-systems
      configuration throughout your project.
    """
    def __init__(self, component, error, message):
        self.component = component
        self.error = error
        self.message = message


class SmifModelRunError(SmifException):
    """Raise when model run requirements are not satisfied
    """
    pass


class SmifValidationError(SmifException):
    """Custom exception to use for parsing validation.
    """
    pass


class SmifTimestepResolutionError(SmifException):
    """Raise when timestep cannot be resolved
    """
    pass



# -*- coding: utf-8 -*-
"""Validate the correct format and presence of the config data
for the system-of-systems model
"""
import itertools

from smif.exception import (SmifDataError, SmifDataInputError,
                            SmifValidationError)

VALIDATION_ERRORS = []


def validate_sos_model_format(sos_model):
    errors = []

    if not isinstance(sos_model, dict):
        msg = "Main config file should contain setup data, instead found: {}"
        err = SmifValidationError(msg.format(sos_model))
        errors.append(err)
        return sos_model

    default_keys = {
        'name': '',
        'description': '',
        'sector_models': [],
        'scenarios': [],
        'narratives': [],
        'model_dependencies': [],
        'scenario_dependencies': []
    }

    # Add default values to missing keys
    for key, value in default_keys.items():
        if key not in sos_model:
            sos_model[key] = value

    # Report keys that should not be in the config
    for key, value in sos_model.items():
        if key not in default_keys:
            errors.append(
                SmifValidationError(
                    'Invalid key `%s` in sos_model configuration `%s`.'
                    % (key, sos_model['name']))
            )

    # Throw collection of errors
    if errors:
        raise SmifDataError(errors)

    return sos_model


def validate_sos_model_config(sos_model, sector_models, scenarios):
    """Check expected values for data loaded from master config file
    """
    errors = []

    if not isinstance(sos_model, dict):
        msg = "Main config file should contain setup data, instead found: {}"
        err = SmifValidationError(msg.format(sos_model))
        errors.append(err)
        return

    # check description
    errors.extend(_validate_description(sos_model))

    # check sector models
    errors.extend(_validate_sos_model_models(sos_model, sector_models))

    # check scenarios
    errors.extend(_validate_sos_model_scenarios(sos_model, scenarios))

    # check narratives
    errors.extend(_validate_sos_model_narratives(sos_model, sector_models))

    # check dependencies
    errors.extend(_validate_sos_model_deps(sos_model, sector_models, scenarios))

    if errors:
        raise SmifDataError(errors)


def _validate_sos_model_models(sos_model, sector_models):
    errors = []
    if not sos_model['sector_models']:
        errors.append(
            SmifDataInputError(
                'sector_models',
                'At least one sector model must be selected.',
                'A system-of-systems model requires to have at least one system ' +
                'enabled to build a valid configuration.'))

    for sector_model in sos_model['sector_models']:
        if sector_model not in [sector_model['name'] for sector_model in sector_models]:
            errors.append(
                SmifDataInputError(
                    'sector_models',
                    '%s must have a valid sector_model configuration.' % (sector_model),
                    'Smif refers to the sector_model-configurations to find ' +
                    'details about a selected sector_model.'))
    return errors


def _validate_sos_model_scenarios(sos_model, scenarios):
    errors = []
    for scenario in sos_model['scenarios']:
        if scenario not in [scenario['name'] for scenario in scenarios]:
            errors.append(
                SmifDataInputError(
                    'scenarios',
                    '%s must have a valid scenario configuration.' % (scenario),
                    'Smif refers to the scenario-configurations to find ' +
                    'details about a selected scenario.'))
    return errors


def _validate_sos_model_narratives(sos_model, sector_models):
    errors = []
    for narrative in sos_model['narratives']:
        # Check provides are valid
        for model_name in narrative['provides']:

            # A narrative can only provides for enabled models
            if model_name not in sos_model['sector_models']:
                errors.append(
                    SmifDataInputError(
                        'narratives',
                        ('Narrative `%s` provides data for model `%s` that is not enabled ' +
                         'in this system-of-systems model.') % (narrative['name'], model_name),
                        'A narrative can only provide for enabled models.'))
            else:
                # A narrative can only provides parameters that exist in the model
                try:
                    sector_model = _pick_sector_model(model_name, sector_models)
                except KeyError:
                    msg = 'Narrative `{}` provides data for model `{}` that is not found.'
                    errors.append(
                        SmifDataInputError(
                            'models',
                            msg.format(narrative['name'], model_name),
                            'A narrative can only provide for existing models.'))
                    sector_model = {'parameters': []}

                parameters = [
                    parameter['name'] for parameter in sector_model['parameters']
                ]
                for provide in narrative['provides'][model_name]:
                    msg = 'Narrative `{}` provides data for non-existing model parameter `{}`'
                    if provide not in parameters:
                        errors.append(
                            SmifDataInputError(
                                'narratives',
                                msg.format(narrative['name'], provide),
                                'A narrative can only provide existing model parameters.'
                            )
                        )

        # Check if all variants are valid
        for variant in narrative['variants']:
            should_provide = list(itertools.chain(*narrative['provides'].values()))
            variant_provides = list(variant['data'].keys())
            if sorted(variant_provides) != sorted(should_provide):
                msg = 'Narrative `{}`, variant `{}` provides incorrect data.'
                errors.append(
                    SmifDataInputError(
                        'narratives',
                        msg.format(narrative['name'], variant['name']),
                        'A variant can only provide data for parameters that are specified ' +
                        'by the narrative.'))
    return errors


def _pick_sector_model(name, models):
    for model in models:
        if model['name'] == name:
            return model
    raise KeyError("Model '{}' not found in models".format(name))


def _validate_sos_model_deps(sos_model, sector_models, scenarios):
    errors = []
    errors.extend(_validate_dependencies(
        sos_model, 'model_dependencies',
        sector_models, 'sector_models',
        sector_models, 'sector_models'
    ))

    errors.extend(_validate_dependencies(
        sos_model, 'scenario_dependencies',
        scenarios, 'scenarios',
        sector_models, 'sector_models'
    ))
    return errors


def _validate_description(configuration):
    errors = []

    if len(configuration['description']) > 255:
        errors.append(
            SmifDataInputError(
                'description',
                'Description must not contain more than 255 characters.',
                'A description should briefly outline a `%s` configuration.'
                % (configuration['name'])))

    return errors


def _validate_dependencies(configuration, conf_key, source, source_key, sink, sink_key):
    errors = []
    for idx, dependency in enumerate(configuration[conf_key]):
        errors.extend(_validate_dependency_cycle(
            idx, dependency, conf_key))
        errors.extend(_validate_dependency_in_sos_model(
            idx, dependency, configuration, conf_key, source_key, sink_key))
        errors.extend(_validate_dependency(
            idx, dependency, conf_key, source, source_key, sink, sink_key))
    return errors


def _validate_dependency_cycle(idx, dependency, conf_key):
    errors = []
    # Circular dependencies are not allowed
    is_current = 'timestep' not in dependency or dependency['timestep'] == 'CURRENT'
    if dependency['source'] == dependency['sink'] and is_current:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Circular dependencies are not allowed.' % (idx + 1),
                'Smif does not support self-dependencies unless the dependency is on ' +
                'output from a previous timestep.'))
    return errors


def _validate_dependency_in_sos_model(idx, dependency, configuration, conf_key, source_key,
                                      sink_key):
    errors = []
    # Source / Sink must be enabled in sos_model config
    if dependency['source'] not in configuration[source_key]:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Source `%s` is not enabled.' %
                (idx + 1, dependency['source']),
                'Each dependency source must be enabled in the sos-model'))

    if dependency['sink'] not in configuration[sink_key]:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Sink `%s` is not enabled.' %
                (idx + 1, dependency['sink']),
                'Each dependency sink must be enabled in the sos-model'))

    # Sink can only have a single dependency
    dep_sinks = [
        (dependency['sink'], dependency['sink_input'])
        for dependency in configuration[conf_key]
    ]
    if dep_sinks.count((dependency['sink'], dependency['sink_input'])) > 1:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Sink input `%s` is driven by multiple sources.'
                % (idx + 1, dependency['sink_input']),
                'A model input can only be driven by a single model output.'))
    return errors


def _validate_dependency(idx, dependency, conf_key, source, source_key, sink,
                         sink_key):
    errors = []
    # Source and sink model configurations must exist
    source_model = [model for model in source if model['name'] == dependency['source']]
    sink_model = [model for model in sink if model['name'] == dependency['sink']]
    if not source_model:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Source `%s` does not exist.' %
                (idx + 1, dependency['source']),
                'Each dependency source must have a `%s` configuration.' %
                (source_key)))
    if not sink_model:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Sink  `%s` does not exist.' %
                (idx + 1, dependency['sink']),
                'Each dependency sink must have a `%s` configuration.' %
                (sink_key)))
    if not sink_model or not source_model:
        # not worth doing further checks if source/sink does not exist
        return errors

    # Source_output and sink_input must exist
    if source_key == 'sector_models':
        source_model_outputs = [
            output for output in source_model[0]['outputs']
            if output['name'] == dependency['source_output']
        ]
    if source_key == 'scenarios':
        source_model_outputs = [
            output for output in source_model[0]['provides']
            if output['name'] == dependency['source_output']
        ]
    sink_model_inputs = [
        input_ for input_ in sink_model[0]['inputs']
        if input_['name'] == dependency['sink_input']
    ]

    if not source_model_outputs:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Source output `%s` does not exist.' %
                (idx + 1, dependency['source_output']),
                'Each dependency source output must exist in the `%s` configuration.' %
                (source_key)))
    if not sink_model_inputs:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Sink input `%s` does not exist.' %
                (idx + 1, dependency['sink_input']),
                'Each dependency sink input must exist in the `%s` configuration.' %
                (sink_key)))
    if not source_model_outputs or not sink_model_inputs:
        # not worth doing further checks if source_output/sink_input does not exist
        return errors

    # Source_output and sink_input must have matching specs
    source_model_output = source_model_outputs[0]
    sink_model_input = sink_model_inputs[0]
    if sorted(source_model_output['dims']) != sorted(sink_model_input['dims']):
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Source `%s` has different dimensions than sink ' % (
                    idx + 1,
                    source_model_output['name']
                ) +
                '`%s` (%s != %s).' % (
                    sink_model_input['name'],
                    source_model_output['dims'],
                    sink_model_input['dims']
                ),
                'Dependencies must have matching dimensions.'))
    if source_model_output['dtype'] != sink_model_input['dtype']:
        errors.append(
            SmifDataInputError(
                conf_key,
                '(Dependency %s) Source `%s` has a different dtype than sink ' % (
                    idx + 1,
                    source_model_output['name'],
                ) +
                '`%s` (%s != %s).' % (
                    sink_model_input['name'],
                    source_model_output['dtype'],
                    sink_model_input['dtype']),
                'Dependencies must have matching data types.'))

    return errors


def validate_path_to_timesteps(timesteps):
    """Check timesteps is a path to timesteps file
    """
    if not isinstance(timesteps, str):
        VALIDATION_ERRORS.append(
            SmifValidationError(
                "Expected 'timesteps' in main config to specify " +
                "a timesteps file, instead got {}.".format(timesteps)))


def validate_timesteps(timesteps, file_path):
    """Check timesteps is a list of integers
    """
    if not isinstance(timesteps, list):
        msg = "Loading {}: expected a list of timesteps.".format(file_path)
        VALIDATION_ERRORS.append(SmifValidationError(msg))
    else:
        msg = "Loading {}: timesteps should be integer years, instead got {}"
        for timestep in timesteps:
            if not isinstance(timestep, int):
                VALIDATION_ERRORS.append(msg.format(file_path, timestep))


def validate_time_intervals(intervals, file_path):
    """Check time intervals
    """
    if not isinstance(intervals, list):
        msg = "Loading {}: expected a list of time intervals.".format(file_path)
        VALIDATION_ERRORS.append(SmifValidationError(msg))
    else:
        for interval in intervals:
            validate_time_interval(interval)


def validate_time_interval(interval):
    """Check a single time interval
    """
    if not isinstance(interval, dict):
        msg = "Expected a time interval, instead got {}.".format(interval)
        VALIDATION_ERRORS.append(SmifValidationError(msg))
        return

    required_keys = ["id", "start", "end"]
    for key in required_keys:
        if key not in interval:
            fmt = "Expected a value for '{}' in each " + \
                "time interval, only received {}"
            VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key, interval)))


def validate_sector_models_initial_config(sector_models):
    """Check list of sector models initial configuration
    """
    if not isinstance(sector_models, list):
        fmt = "Expected 'sector_models' in main config to " + \
              "specify a list of sector models to run, instead got {}."
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(sector_models)))
    else:
        if len(sector_models) == 0:
            VALIDATION_ERRORS.append(
                SmifValidationError("No 'sector_models' specified in main config file."))

        # check each sector model
        for sector_model_config in sector_models:
            validate_sector_model_initial_config(sector_model_config)


def validate_sector_model_initial_config(sector_model_config):
    """Check a single sector model initial configuration
    """
    if not isinstance(sector_model_config, dict):
        fmt = "Expected a sector model config block, instead got {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(sector_model_config)))
        return

    required_keys = ["name", "config_dir", "path", "classname"]
    for key in required_keys:
        if key not in sector_model_config:
            fmt = "Expected a value for '{}' in each " + \
                  "sector model in main config file, only received {}"
            VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key, sector_model_config)))


def validate_dependency_spec(input_spec, model_name):
    """Check the input specification for a single sector model
    """
    if not isinstance(input_spec, list):
        fmt = "Expected a list of parameter definitions in '{}' model " + \
              "input specification, instead got {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(model_name, input_spec)))
        return

    for dep in input_spec:
        validate_dependency(dep)


def validate_dependency(dep):
    """Check a dependency specification
    """
    if not isinstance(dep, dict):
        fmt = "Expected a dependency specification, instead got {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(dep)))
        return

    required_keys = ["name", "spatial_resolution", "temporal_resolution", "units"]
    for key in required_keys:
        if key not in dep:
            fmt = "Expected a value for '{}' in each model dependency, only received {}"
            VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key, dep)))


def validate_scenario_data_config(scenario_data):
    """Check scenario data
    """
    if not isinstance(scenario_data, list):
        fmt = "Expected a list of scenario datasets in main model config, " + \
              "instead got {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(scenario_data)))
        return

    for scenario in scenario_data:
        validate_scenario(scenario)


def validate_scenario(scenario):
    """Check a single scenario specification
    """
    if not isinstance(scenario, dict):
        fmt = "Expected a scenario specification, instead got {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(scenario)))
        return

    required_keys = ["parameter", "spatial_resolution", "temporal_resolution", "units", "file"]
    for key in required_keys:
        if key not in scenario:
            fmt = "Expected a value for '{}' in each scenario, only received {}"
            VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key, scenario)))


def validate_scenario_data(data, file_path):
    """Check a list of scenario observations
    """
    if not isinstance(data, list):
        fmt = "Expected a list of scenario data in {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(file_path)))
        return

    for datum in data:
        validate_scenario_datum(datum, file_path)


def validate_scenario_datum(datum, file_path):
    """Check a single scenario datum
    """
    if not isinstance(datum, dict):
        fmt = "Expected a scenario data point, instead got {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(datum)))
        return

    required_keys = ["region", "interval", "year", "value"]
    for key in required_keys:
        if key not in datum:
            fmt = "Expected a value for '{}' in each data point in a scenario, " + \
                  "only received {}"
            VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key, datum)))


def validate_initial_conditions(data, file_path):
    """Check a list of initial condition observations
    """
    if not isinstance(data, list):
        fmt = "Expected a list of initial conditions in {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(file_path)))
        return

    for datum in data:
        validate_initial_condition(datum, file_path)


def validate_initial_condition(datum, file_path):
    """Check a single initial condition datum
    """
    if not isinstance(datum, dict):
        fmt = "Expected a initial condition data point, instead got {} from {}"
        VALIDATION_ERRORS.append(SmifValidationError(fmt.format(datum, file_path)))
        return

    required_keys = ["name", "build_date"]
    for key in required_keys:
        if key not in datum:
            fmt = "Expected a value for '{}' in each data point in a initial condition, " + \
                  "only received {} from {}"
            VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key, datum, file_path)))


def validate_planning_config(planning):
    """Check planning options
    """
    required_keys = ["pre_specified", "rule_based", "optimisation"]
    for key in required_keys:
        if key not in planning:
            fmt = "No '{}' settings specified under 'planning' " + \
                  "in main config file."
            VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key)))

    # check each planning type
    for key, planning_type in planning.items():
        if "use" not in planning_type:
            fmt = "No 'use' settings specified for '{}' 'planning'"
            VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key)))
            continue
        if planning_type["use"]:
            if "files" not in planning_type or \
               not isinstance(planning_type["files"], list) or \
               len(planning_type["files"]) == 0:

                fmt = "No 'files' provided for the '{}' " + \
                      "planning type in main config file."
                VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key)))


def validate_region_sets_config(region_sets):
    """Check regions sets
    """
    required_keys = ["name", "file"]
    for key in required_keys:
        for region_set in region_sets:
            if key not in region_set:
                fmt = "Expected a value for '{}' in each " + \
                    "region set in main config file, only received {}"
                VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key, region_set)))


def validate_interval_sets_config(interval_sets):
    """Check interval sets
    """
    required_keys = ["name", "file"]
    for key in required_keys:
        for interval_set in interval_sets:
            if key not in interval_set:
                fmt = "Expected a value for '{}' in each " + \
                    "interval set in main config file, only received {}"
                VALIDATION_ERRORS.append(SmifValidationError(fmt.format(key, interval_set)))


def validate_interventions(data, path):
    """Validate the loaded data as required for model interventions
    """
    # check required keys
    required_keys = ["name", "location", "capital_cost", "operational_lifetime",
                     "economic_lifetime"]

    # except for some keys which are allowed simple values,
    # expect each attribute to be of the form {value: x, units: y}
    simple_keys = ["name", "sector", "location"]

    for intervention in data:
        for key in required_keys:
            if key not in intervention:
                fmt = "Loading interventions from {}, required " + \
                      "a value for '{}' in each intervention, but only " + \
                      "received {}"
                VALIDATION_ERRORS.append(
                    SmifValidationError(fmt.format(path, key, intervention)))

        for key, value in intervention.items():
            if key not in simple_keys and (
                    not isinstance(value, dict)
                    or "value" not in value
                    or "units" not in value):
                fmt = "Loading interventions from {3}, {0}.{1} was {2} but " + \
                      "should have specified units, " + \
                      "e.g. {{'value': {2}, 'units': 'm'}}"

                msg = fmt.format(intervention["name"], key, value, path)
                VALIDATION_ERRORS.append(SmifValidationError(msg))



"""Test config validation
"""
from smif.data_layer.validate import validate_sos_model_config
from smif.exception import SmifDataError


class TestValidateSosModel:
    """Check that validation raises validation errors when one part of the
    configuration is incorrect
    """
    def test_correct(self, get_sos_model, get_sector_model, energy_supply_sector_model,
                     sample_scenarios):
        """Expect no error on the default configuration
        """
        validate_sos_model_config(
            get_sos_model,
            [get_sector_model, energy_supply_sector_model],
            sample_scenarios)

    def test_description_too_long(self, get_sos_model, get_sector_model,
                                  energy_supply_sector_model, sample_scenarios):
        """Expect exception when description is too long
        """
        get_sos_model['description'] = 255 * 'a'
        validate_sos_model_config(
            get_sos_model,
            [get_sector_model, energy_supply_sector_model],
            sample_scenarios)

        try:
            get_sos_model['description'] = 256 * 'a'
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'description'
            assert 'characters' in ex.args[0][0].error

    def test_sector_models_none_configured(self, get_sos_model, get_sector_model,
                                           energy_supply_sector_model, sample_scenarios):
        """Expect exception when no sector_models are configured
        """
        try:
            get_sos_model['sector_models'] = []
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'sector_models'
            assert 'one sector model must be selected' in ex.args[0][0].error

    def test_sector_model_missing_reference(self, get_sos_model, sample_scenarios):
        """Expect error when references sector model
        configuration does not exist
        """
        try:
            validate_sos_model_config(
                get_sos_model,
                [],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'sector_models'
            assert 'valid sector_model configuration' in ex.args[0][0].error

    def test_scenario_missing_reference(self, get_sos_model, get_sector_model,
                                        energy_supply_sector_model, sample_scenarios):
        """Expect error when references scenario
        configuration does not exist
        """
        try:
            get_sos_model['scenarios'] = ['scenario_a']
            sample_scenarios = []
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'scenarios'
            assert 'valid scenario configuration' in ex.args[0][0].error

    def test_dependencies_circular(self, get_sos_model, get_sector_model,
                                   energy_supply_sector_model, sample_scenarios):
        """Expect error when circular dependency is in
        configuration
        """
        get_sos_model['sector_models'] = [
            'energy_demand'
        ]
        get_sos_model['model_dependencies'] = [
            {
                "source": "energy_demand",
                "source_output": "output_a",
                "sink": "energy_demand",
                "sink_input": "input_a"
            }
        ]
        get_sector_model['inputs'] = [
            {
                "name": "input_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]
        get_sector_model['outputs'] = [
            {
                "name": "output_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]

        try:
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'model_dependencies'
            assert 'Circular dependencies' in ex.args[0][0].error

    def test_dependencies_source_or_sink_not_enabled(self, get_sos_model, get_sector_model,
                                                     energy_supply_sector_model,
                                                     sample_scenarios):
        """Expect error when source or sink is not enabled
        in the sector_model configuration
        """
        get_sos_model['model_dependencies'] = [
            {
                "source": "energy_demand",
                "source_output": "output_a",
                "sink": "energy_supply",
                "sink_input": "input_a"
            }
        ]
        get_sos_model['sector_models'] = [
            'energy_supply'
        ]
        get_sector_model['outputs'] = [
            {
                "name": "output_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]
        energy_supply_sector_model['inputs'] = [
            {
                "name": "input_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]

        try:
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][2].component == 'model_dependencies'
            assert '`energy_demand` is not enabled' in ex.args[0][2].error

    def test_dependencies_source_output_or_sink_input_not_exist(
            self, get_sos_model, get_sector_model, energy_supply_sector_model,
            sample_scenarios):
        """Expect error when source output or sink input do not
        exist
        """
        get_sos_model['model_dependencies'] = [
            {
                "source": "energy_demand",
                "source_output": "output_a",
                "sink": "energy_supply",
                "sink_input": "input_a"
            }
        ]
        get_sector_model['outputs'] = []
        energy_supply_sector_model['inputs'] = [
            {
                "name": "input_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]

        try:
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'model_dependencies'
            assert 'Source output `output_a` does not exist' in ex.args[0][0].error

        get_sector_model['outputs'] = [
            {
                "name": "output_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]
        energy_supply_sector_model['inputs'] = []

        try:
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'model_dependencies'
            assert 'Sink input `input_a` does not exist' in ex.args[0][0].error

    def test_dependencies_has_matching_specs(self, get_sos_model, get_sector_model,
                                             energy_supply_sector_model, sample_scenarios):
        """Expect error when source output or sink input have
        different specs
        """
        get_sos_model['model_dependencies'] = [
            {
                "source": "energy_demand",
                "source_output": "output_a",
                "sink": "energy_supply",
                "sink_input": "input_a"
            }
        ]
        get_sector_model['outputs'] = [
            {
                "name": "output_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]
        energy_supply_sector_model['inputs'] = [
            {
                "name": "input_a",
                "dims": [
                    "dim_b"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]

        try:
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'model_dependencies'
            msg = '`output_a` has different dimensions than sink `input_a`'
            assert msg in ex.args[0][0].error

        get_sector_model['outputs'] = [
            {
                "name": "output_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]
        energy_supply_sector_model['inputs'] = [
            {
                "name": "input_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_b",
                "unit": "unit_a"
            }
        ]

        try:
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert ex.args[0]
            assert ex.args[0][0].component == 'model_dependencies'
            msg = '`output_a` has a different dtype than sink `input_a`'
            assert msg in ex.args[0][0].error

    def test_dependencies_sink_input_can_only_be_driven_once(
            self, get_sos_model, get_sector_model, energy_supply_sector_model,
            sample_scenarios):
        """Expect error when sink input is driven by multiple
        sources
        """
        get_sos_model['model_dependencies'] = [
            {
                "source": "energy_demand",
                "source_output": "output_a",
                "sink": "energy_supply",
                "sink_input": "input_a"
            },
            {
                "source": "energy_demand",
                "source_output": "output_b",
                "sink": "energy_supply",
                "sink_input": "input_a"
            }
        ]
        get_sector_model['outputs'] = [
            {
                "name": "output_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            },
            {
                "name": "output_b",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]
        energy_supply_sector_model['inputs'] = [
            {
                "name": "input_a",
                "dims": [
                    "dim_a"
                ],
                "dtype": "dtype_a",
                "unit": "unit_a"
            }
        ]

        try:
            validate_sos_model_config(
                get_sos_model,
                [get_sector_model, energy_supply_sector_model],
                sample_scenarios)
            assert False
        except SmifDataError as ex:
            assert len(ex.args[0]) == 2
            assert ex.args[0][0].component == 'model_dependencies'
            assert 'Sink input `input_a` is driven by multiple sources' in ex.args[0][0].error
            assert ex.args[0][1].component == 'model_dependencies'
            assert 'Sink input `input_a` is driven by multiple sources' in ex.args[0][0].error



"""Tests functionality of NDimensionalRegister class that computes coefficients for
different operations
"""
import numpy as np
from pytest import mark
from smif.convert.adaptor import Adaptor


class TestPerformConversion:
    """Convert between dimension resolutions
    """
    @mark.parametrize('space, time, expected', [
        # Space disaggregation only
        (
            np.array([[0.333, 0.333, 0.333]]),
            np.array([[1.0]]),
            np.array([[0.333],
                      [0.333],
                      [0.333]])
        ),
        # Time disaggregation only
        (
            np.array([[1.0]]),
            np.array([[0.333, 0.333, 0.333]]),
            np.array([[0.333, 0.333, 0.333]])
        ),
        # Space and time disaggregation
        (
            np.array([[0.333333, 0.333333, 0.333333]]),
            np.array([[0.333333, 0.333333, 0.333333]]),
            np.array([[0.111, 0.111, 0.111],
                      [0.111, 0.111, 0.111],
                      [0.111, 0.111, 0.111]])
        )
    ])
    def test_disaggregation_operation(self, space, time, expected):
        """Disggregations should split data along the transformed dimension
        """
        data = np.array([[1]])

        intermediate = Adaptor.convert_with_coefficients(data, space, 0)
        actual = Adaptor.convert_with_coefficients(intermediate, time, 1)
        np.testing.assert_allclose(actual, expected, rtol=1e-2)

    @mark.parametrize('space, time, expected', [
        # Space aggregation only
        (
            np.array([[1],
                      [1]]),
            np.array([[1, 0, 0],
                      [0, 1, 0],
                      [0, 0, 1]]),
            np.array([[666.666, 666.666, 666.666]]),
        ),
        # Time aggregation only
        (
            np.array([[1, 0],
                      [0, 1]]),
            np.array([[1],
                      [1],
                      [1]]),
            np.array([[1000],
                      [1000]])
        ),
        # Space and time aggregation
        (
            np.array([[1],
                      [1]]),
            np.array([[1],
                      [1],
                      [1]]),
            np.array([[2000]]))
    ])
    def test_aggregation_operation(self, space, time, expected):
        """Aggregations should collect data along the transformed dimension
        """
        # Two regions, three intervals
        data = np.array([[333.333, 333.333, 333.333],
                         [333.333, 333.333, 333.333]])
        intermediate = Adaptor.convert_with_coefficients(data, space, 0)
        actual = Adaptor.convert_with_coefficients(intermediate, time, 1)
        np.testing.assert_allclose(actual, expected, rtol=1e-2)

    def test_multidimensional_operation(self):
        """Operations over 3-dimensional data
        """
        # start with something (1, 2, 3)
        data = np.array(
            [[[0.0, 1.0, 2.0],
              [3.0, 4.0, 5.0]]]
        )

        # split 1st dim (2, 2, 3)
        coefficients = np.ones((1, 2)) / 2
        expected = np.array(
            [[[0.0, 0.5, 1.0],
              [1.5, 2.0, 2.5]],

             [[0.0, 0.5, 1.0],
              [1.5, 2.0, 2.5]]]
        )
        actual = Adaptor.convert_with_coefficients(data, coefficients, 0)
        np.testing.assert_allclose(actual, expected)

        # sum 3rd dim (2, 2, 1)
        coefficients = np.ones((3, 1))
        expected = np.array(
            [[[1.5],
              [6.0]],

             [[1.5],
              [6.0]]]
        )
        actual = Adaptor.convert_with_coefficients(actual, coefficients, 2)
        np.testing.assert_allclose(actual, expected)



import logging
import os
import sys
import traceback

from smif.controller.modelrun import ModelRunBuilder
from smif.data_layer.model_loader import ModelLoader
from smif.exception import SmifDataNotFoundError
from smif.model import ScenarioModel, SosModel

LOGGER = logging.getLogger(__name__)


def get_model_run_definition(store, modelrun):
    """Builds the model run

    Arguments
    ---------
    store : ~smif.data_layer.store.Store
        Path to the project directory
    modelrun : str
        Name of the model run to run

    Returns
    -------
    dict
        The complete model_run configuration dictionary with contained
        ScenarioModel, SosModel and SectorModel objects

    """
    try:
        model_run_config = store.read_model_run(modelrun)
    except SmifDataNotFoundError:
        LOGGER.error("Model run %s not found. Run 'smif list' to see available model runs.",
                     modelrun)
        exit(-1)

    LOGGER.info("Running %s", model_run_config['name'])
    LOGGER.debug("Model Run: %s", model_run_config)
    sos_model_config = store.read_sos_model(model_run_config['sos_model'])

    sector_models = get_sector_models(sos_model_config['sector_models'], store)
    LOGGER.debug("Sector models: %s", sector_models)

    scenario_models = get_scenario_models(model_run_config['scenarios'], store)
    LOGGER.debug("Scenario models: %s", [model.name for model in scenario_models])

    sos_model = SosModel.from_dict(sos_model_config, sector_models + scenario_models)
    model_run_config['sos_model'] = sos_model
    LOGGER.debug("Model list: %s", list(model.name for model in sos_model.models))

    model_run_config['strategies'] = store.read_strategies(model_run_config['name'])
    LOGGER.debug("Strategies: %s", [s['type'] for s in model_run_config['strategies']])

    return model_run_config


def get_scenario_models(scenarios, handler):
    """Read in ScenarioModels

    Arguments
    ---------
    scenarios : dict
    handler : smif.data_layer.Store

    Returns
    -------
    list of ScenarioModel
    """
    scenario_models = []
    for scenario_name, variant_name in scenarios.items():
        scenario_definition = handler.read_scenario(scenario_name)

        # assign variant name to definition
        scenario_definition['scenario'] = variant_name

        # rename provides => outputs
        scenario_definition['outputs'] = scenario_definition['provides']
        del scenario_definition['provides']

        LOGGER.debug("Scenario definition: %s", scenario_name)

        scenario_model = ScenarioModel.from_dict(scenario_definition)
        scenario_models.append(scenario_model)
    return scenario_models


def get_sector_models(sector_model_names, handler):
    """Read and build SectorModels

    Arguments
    ---------
    sector_model_names : list of str
    handler : smif.data_layer.Store

    Returns
    -------
    list of SectorModel implementations
    """
    sector_models = []
    loader = ModelLoader()
    for sector_model_name in sector_model_names:
        sector_model_config = handler.read_model(sector_model_name)

        # absolute path to be crystal clear for ModelLoader when loading python class
        sector_model_config['path'] = os.path.normpath(
            os.path.join(handler.model_base_folder, sector_model_config['path'])
        )
        sector_model = loader.load(sector_model_config)
        sector_models.append(sector_model)
    return sector_models


def build_model_run(model_run_config):
    """Builds the model run

    Arguments
    ---------
    model_run_config: dict
        A valid model run configuration dict with objects

    Returns
    -------
    `smif.controller.modelrun.ModelRun`
    """
    LOGGER.profiling_start('build_model_run', model_run_config['name'])
    try:
        builder = ModelRunBuilder()
        builder.construct(model_run_config)
        modelrun = builder.finish()
    except AssertionError as error:
        err_type, err_value, err_traceback = sys.exc_info()
        traceback.print_exception(err_type, err_value, err_traceback)
        err_msg = str(error)
        if err_msg:
            LOGGER.error("An AssertionError occurred (%s) see details above.", err_msg)
        else:
            LOGGER.error("An AssertionError occurred, see details above.")
        exit(-1)

    LOGGER.profiling_stop('build_model_run', model_run_config['name'])
    return modelrun
